""" aparse_types.py -- Data types useful for processing parsed action language """

from collections import namedtuple
from typing import NamedTuple
from enum import Enum


class MaxMult(Enum):
    ONE = 1
    MANY = 2


class Content(Enum):
    SCALAR = 1
    INSTANCE = 2
    RELATION = 3

Attribute_ap = namedtuple('Attribute_ap', 'name scalar')
""" An Attribute/Scalar pair """
Labeled_Flow = namedtuple('Labeled_Flow', 'label flow')
""" A label, Flow_ap pair """
Attribute_Comparison = namedtuple('Attribute_Comparison', 'attr op')
""" An attribute compared in a selection phrase """
Boundary_Actions = namedtuple("Boundary_Actions", "ain aout")
""" Initial actions not dependent on any data flow input and output actions that do not flow to any other action"""
Activity_ap = namedtuple("Activity_ap", "anum domain cname sname eename opname xiflow activity_path scrall_text")
""" Activity identification and diagnostic data """

class Flow_ap(NamedTuple):
    """
    Describes a generated flow
    """
    fid: str  # Flow ID
    content: Content
    tname: str
    max_mult: MaxMult


"""
relationship.py – Convert parsed relationship to a relation
"""

import logging
from xuml_populate.config import mmdb
from xuml_populate.populate.element import Element
from xuml_populate.populate.generalization import Generalization
from xuml_populate.populate.binary_association import BinaryAssociation
from xuml_populate.populate.ordinal import Ordinal
from pyral.transaction import Transaction
from pyral.relvar import Relvar
from xuml_populate.exceptions.mp_exceptions import UnknownRelationshipType
from xuml_populate.populate.mmclass_nt import Relationship_i

_logger = logging.getLogger(__name__)

# Transactions
tr_Rel = "Relationship"

class Relationship:
    """
    Populate a Relationship
    """
    rnum = None

    @classmethod
    def populate(cls, domain: str, subsystem, record):
        """
        Populate all relevant Relationship relvars

        :param domain: The domain name
        :param subsystem: The subsystem name
        :param record: Parse of the relationship
        """

        cls.rnum = record['rnum']

        # Populate relationship
        Transaction.open(mmdb, tr_Rel)

        Element.populate_labeled_subys_element(tr=tr_Rel, label=cls.rnum, subsystem=subsystem.name, domain=domain)
        Relvar.insert(mmdb, tr=tr_Rel, relvar='Relationship', tuples=[
            Relationship_i(Rnum=cls.rnum, Domain=domain)
        ])

        # Populate based on relationship type
        if 't_side' in record:
            BinaryAssociation(tr=tr_Rel, domain=domain, rnum=cls.rnum, record=record)
        elif 'superclass' in record:
            Generalization(tr=tr_Rel, domain=domain, rnum=cls.rnum, record=record)
        elif 'ascend' in record:
            Ordinal(tr=tr_Rel, domain=domain, rnum=cls.rnum, record=record)
        else:
            _logger.error(
                "Population encountered relationship type that is not an Association, Generalization, or Ordinal.")
            raise UnknownRelationshipType
        Transaction.execute(db=mmdb, name=tr_Rel)



"""
read_action.py – Populate a read action instance in PyRAL
"""

import logging
from typing import Set, List, Tuple
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content, Activity_ap
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.mm_class import MMclass
from scrall.parse.visitor import Projection_a
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.mmclass_nt import Read_Action_i, Attribute_Read_Access_i
from pyral.relvar import Relvar
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

# Transactions
tr_Read = "Read Action"

class ReadAction:
    """
    Create all relations for a Read Action
    """

    input_instance_flow = None  # We are selecting instances from this instance flow
    output_instance_flow = None
    expression = None
    comparison_criteria = []
    equivalence_criteria = []
    restriction_text = ""
    cardinality = None
    action_id = None
    criterion_ctr = 0
    max_mult = None

    @classmethod
    def populate(cls, input_single_instance_flow: Flow_ap, attrs: Tuple[str],
                 anum: str, domain: str) -> (str, Tuple[Flow_ap]):
        """
        Populate the Read Action

        :param input_single_instance_flow: The summary of the input flow to the Read Action
        :param attrs: A tuple of attribute names to read
        :param anum:  The activity number
        :param domain:  The domain name
        :return: A tuple of scalar flows matching the order of the specified attrs
        """
        # TODO: Avoid populating duplicate for reading the same attr set by maintaining dict of action/flows
        # TODO: for the current activity. If duplicate found, return its action id and output flows instead.
        # TODO: Ensure that the existing read action that has no incoming control flows

        assert input_single_instance_flow.content == Content.INSTANCE
        assert input_single_instance_flow.max_mult == MaxMult.ONE
        cname = input_single_instance_flow.tname

        # Get the class header
        class_attrs = MMclass.header(cname=cname, domain=domain)

        # Populate the Action superclass instance and obtain its action_id
        Transaction.open(db=mmdb, name=tr_Read)
        action_id = Action.populate(tr=tr_Read, anum=anum, domain=domain, action_type="read")  # Transaction open
        Relvar.insert(db=mmdb, tr=tr_Read, relvar='Read_Action', tuples=[
            Read_Action_i(ID=action_id, Activity=anum, Domain=domain, Instance_flow=input_single_instance_flow.fid)
        ])
        scalar_flows = []
        for a in attrs:
            of = Flow.populate_scalar_flow(scalar_type=class_attrs[a], anum=anum, domain=domain, label=None)
            Relvar.insert(db=mmdb, tr=tr_Read, relvar='Attribute_Read_Access', tuples=[
                Attribute_Read_Access_i(Attribute=a, Class=cname, Read_action=action_id, Activity=anum,
                                        Domain=domain, Output_flow=of.fid)
            ])
            scalar_flows.append(of)

            # output_flows[pa] = of
        # We now have a transaction with all select-action instances, enter into the metamodel db
        Transaction.execute(db=mmdb, name=tr_Read)  # Select Action
        return action_id, tuple(scalar_flows)



"""
element.py – Populate an element instance in PyRAL
"""

import logging
from pyral.relvar import Relvar
from xuml_populate.config import mmdb
from xuml_populate.populate.mmclass_nt import Element_i, Spanning_Element_i, Subsystem_Element_i
# TODO: Add spanning element support

_signum_counters = {}  # A separate counter per domain
_logger = logging.getLogger(__name__)

class Element:
    """
    Create a State Model relation
    """
    _num_counters = {}  # A separate number counter per domain

    @classmethod
    def init_counter(cls, key: str) -> int:
        """
        Create a new counter using the supplied key if it does not already exist in the num counter dict

        :param key: Usually the domain name but could be subsys:domain or something else
        :return: The next available number
        """
        # Should refactor this into an Element population numbering method
        if key not in cls._num_counters:
            cls._num_counters[key] = 1
        else:
            cls._num_counters[key] += 1
        return cls._num_counters[key]

    @classmethod
    def populate_unlabeled_subsys_element(cls, tr: str, prefix: str, subsystem: str, domain: str) -> str:
        """
        Generates a label for a new Subsystem Element and populates it

        :param tr:  An open transaction name
        :param prefix: Prefixed to counter to create unique string label
        :param subsystem: The name of the subsystem since these are Subsystem Elements
        :param domain: The element belongs to this domain
        :return: generated label such as SIG12, A47, etc
        """

        label = f'{prefix}{cls.init_counter(key=domain)}'
        Relvar.insert(mmdb, tr=tr, relvar='Element', tuples=[
            Element_i(Label=label, Domain=domain)
        ])
        Relvar.insert(mmdb, tr=tr, relvar='Subsystem_Element', tuples=[
            Subsystem_Element_i(Label=label, Domain=domain, Subsystem=subsystem)
        ])
        return label

    @classmethod
    def populate_labeled_subys_element(cls, tr: str, label: str, subsystem: str, domain: str):
        """
        Populates pre-labeled Subsystem Element such as cnum and rnum

        Rnums are typically specified and remembered by the user directly in the xcm files
        (or in whatever tool they are using to specify a class model as they develop).

        Most users ignore the Cnums so they can be generated behind the scenes and kept as an internal labeling
        system for the most part. They are not specified in the xcm files for now (but may be later)

        :param tr:  An open transaction name
        :param label: The user or generated label such as R812 for rnums or C7 for cnums
        :param subsystem: The name of the subsystem since these are Subsystem Elements
        :param domain: The element belongs to this domain
        """

        # We don't need to use our counter since the label has already been specified
        # so we just insert the Element and Subystem Element classes of the Domain Subsystem
        Relvar.insert(mmdb, tr=tr, relvar='Element', tuples=[
            Element_i(Label=label, Domain=domain)
        ])
        Relvar.insert(mmdb, tr=tr, relvar='Subsystem_Element', tuples=[
            Subsystem_Element_i(Label=label, Domain=domain, Subsystem=subsystem)
        ])


    @classmethod
    def depopulate_spanning_element(cls, tr: str, label: str, domain: str):
        """
        Remove the specified spanning elmenent type from the database.

        Assumes that the spanning element subclass is managing the db transaction

        :param tr:  An open transaction name
        :param label: The user or generated label such as R812 for rnums or C7 for cnums
        :param domain: The element belongs to this domain
        """
        Relvar.deleteone(mmdb, tr=tr, relvar_name='Spanning_Element', tid={'Label': label, 'Domain': domain})
        Relvar.deleteone(mmdb, tr=tr, relvar_name='Element', tid={'Label': label, 'Domain': domain})


""" method_activity.py -- Process all execution components of a method's activity """

# System
import logging
from typing import NamedTuple

# Model Integration
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction
from pyral.rtypes import JoinCmd, ProjectCmd, SetCompareCmd, SetOp, Attribute, SumExpr, RelationValue

# xUML Populate
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.aparse_types import Activity_ap
from xuml_populate.populate.xunit import ExecutionUnit
from xuml_populate.populate.mmclass_nt import Flow_Dependency_i, Wave_i, Wave_Assignment_i
from xuml_populate.exceptions.action_exceptions import MethodXIFlowNotPopulated

_logger = logging.getLogger(__name__)

# TODO: This can be generated later by make_repo, ensure each name ends with 'Action'
class UsageAttrs(NamedTuple):
    cname: str
    id_attr: str | None
    in_attr: str | None
    out_attr: str | None


flow_attrs = [
    UsageAttrs(cname='Select_Action', id_attr='ID', in_attr='Input_flow', out_attr=None),
    UsageAttrs(cname='Traverse_Action', id_attr='ID', in_attr='Source_flow', out_attr='Destination_flow'),
    UsageAttrs(cname='Many_Select', id_attr='ID', in_attr=None, out_attr='Output_flow'),
    UsageAttrs(cname='Single_Select', id_attr='ID', in_attr=None, out_attr='Output_flow'),
    UsageAttrs(cname='Table_Action', id_attr='ID', in_attr='Input_a_flow', out_attr='Output_flow'),
    UsageAttrs(cname='Set_Action', id_attr='ID', in_attr='Input_b_flow', out_attr=None),
    UsageAttrs(cname='Read_Action', id_attr='ID', in_attr='Instance_flow', out_attr=None),
    UsageAttrs(cname='Attribute_Read_Access', id_attr='Read_action', in_attr=None, out_attr='Output_flow'),
    UsageAttrs(cname='Comparison_Criterion', id_attr='Action', in_attr='Value', out_attr=None),
    UsageAttrs(cname='Gate_Action', id_attr='ID', in_attr=None, out_attr='Output_flow'),
    UsageAttrs(cname='Gate_Input', id_attr='Gate_action', in_attr='Input_flow', out_attr=None),
    UsageAttrs(cname='Case', id_attr='Switch_action', in_attr=None, out_attr='Flow'),
    UsageAttrs(cname='Control_Dependency', id_attr='Action', in_attr='Control_flow', out_attr=None),
    UsageAttrs(cname='Extract_Action', id_attr='ID', in_attr='Input_tuple', out_attr='Output_scalar'),
]

# For debugging
from collections import namedtuple
Action_i = namedtuple('Action_i', 'ID')

class MethodActivity:

    def __init__(self, name: str, class_name: str, method_data, activity_data, domain: str):

        self.name = name
        self.class_name = class_name
        self.activity_data = activity_data
        self.anum = self.activity_data['anum']
        self.method_data = method_data
        self.domain = domain
        self.wave_ctr = 1  # Initialize wave counter
        self.waves = dict()
        self.xactions = None

        self.pop_xunits()
        self.pop_flow_dependencies()
        self.assign_waves()
        self.populate_waves()

        # Now we can populate the Wave and Wave Assignment relvars
        pass

    def populate_waves(self):
        """
        Populate Wave and Wave Assignment class relvars
        """
        for wave_number, wave_actions in self.waves.items():
            wave_tr = f"wave_{wave_number}"
            Transaction.open(db=mmdb, name=wave_tr)
            Relvar.insert(db=mmdb, relvar="Wave", tuples=[
                Wave_i(Number=wave_number, Activity=self.anum, Domain=self.domain)
            ], tr=wave_tr)
            wa_tuples = [Wave_Assignment_i(Action=a, Activity=self.anum, Domain=self.domain, Wave=wave_number)
                         for a in wave_actions]
            Relvar.insert(db=mmdb, relvar="Wave_Assignment", tuples=wa_tuples, tr=wave_tr)
            Transaction.execute(db=mmdb, name=wave_tr)
        # Relation.print(db=mmdb, variable_name="Wave")
        # Relation.print(db=mmdb, variable_name="Wave_Assignment")

    def initially_enabled_flows(self):
        """
        For a Method Activity, there are three kinds of Flows that are enabled prior to any Action execution.
        In other words, these Flows are enabled before the first Wave executes.

        These are all parameter flows, the executing instance flow, and the optional output flow

        Here we create a relation representing the set of all flow IDs for these flows.
        """
        # Get the executing instance flow (probably always F1, but just to be safe...)
        # It's a referential attribute of Method, so we just project on that

        # Get the method for our Anum and Domain
        R = f"Anum:<{self.activity_data['anum']}>, Domain:<{self.domain}>"
        Relation.restrict(db=mmdb, relation='Method', restriction=R)
        # Project on the referential attribute to the flow
        Relation.project(db=mmdb, attributes=("Executing_instance_flow",))
        # Rename it to "Flow"
        Relation.rename(db=mmdb, names={"Executing_instance_flow": "Flow"}, svar_name="xi_flow")
        # Relation.print(db=mmdb, variable_name="xi_flow")

        # Get the parameter input flows for our Anum and Domain
        R = f"Activity:<{self.activity_data['anum']}>, Domain:<{self.domain}>"
        Relation.restrict(db=mmdb, relation='Parameter', restriction=R)
        # Project on the referential attribute to the flow
        Relation.project(db=mmdb, attributes=("Input_flow",))
        # Rename it to "Flow"
        Relation.rename(db=mmdb, names={"Input_flow": "Flow"}, svar_name="param_flows")
        # Relation.print(db=mmdb, variable_name="param_flows")

        # Get any class accessor flows for our Anum and Domain
        R = f"Activity:<{self.activity_data['anum']}>, Domain:<{self.domain}>"
        Relation.restrict(db=mmdb, relation='Class_Accessor', restriction=R)
        # Project on the referential attribute to the flow
        Relation.project(db=mmdb, attributes=("Output_flow",))
        # Rename it to "Flow"
        Relation.rename(db=mmdb, names={"Output_flow": "Flow"}, svar_name="class_flows")
        # Relation.print(db=mmdb, variable_name="class_flows")

        # Now take the union of all three
        Relation.union(db=mmdb, relations=("xi_flow", "param_flows", "class_flows"), svar_name="wave_enabled_flows")
        # Relation.print(db=mmdb, variable_name="wave_enabled_flows")

    def initially_executable_actions(self) -> RelationValue:
        """
        Identify all actions that can be executed in the first wave.

        To populate the first wave of execution we need to gather all Actions whose inputs consist entirely
        of Flows that are initially available. In other words, each Action in the initial Wave will require
        no inputs from any other Action.
        """
        # Get all downstream actions (actions that require input from other actions)

        # First we'll need to pair down the Flow Dependency instances to just those
        # in our Method Activity
        R = f"Activity:<{self.activity_data['anum']}>, Domain:<{self.domain}>"
        Relation.restrict(db=mmdb, relation='Flow_Dependency', restriction=R, svar_name="fd")
        # fd relaton set to local Flow Dependency instances

        # Define downstream_actions as the ID's of those Actions that take at least one input from some other Action
        Relation.project(db=mmdb, attributes=("To_action",))  # Actions that are a Flow destination
        Relation.rename(db=mmdb, names={"To_action": "ID"}, svar_name="downstream_actions")

        # Now subtract those downstream actions from the set of unexecuted_actions (all Actions at this point)
        ia = Relation.subtract(db=mmdb, rname1="unexecuted_actions", rname2="downstream_actions", svar_name="executable_actions")
        # Relation.print(db=mmdb, variable_name="executable_actions")
        # Here we are initializing a relation variable named "executable_actions" that will update for each new Wave
        # And in Wave 1, it is the set of initial actions
        return ia

    def enable_action_outputs(self, source_action_relation: str):
        """
        Given a set of Action IDs as a relation value, get the set of all output Flows for those Actions
        Those Flows are saved in the wave_enabled_flows relation value

        :param source_action_relation: This relation is a set of Action IDs representing the source of the output Flows
        """
        # Lookup all Flow Dependencies where the source action is a From_action and collect the output Flows
        Relation.join(db=mmdb, rname1=source_action_relation, rname2="fd", attrs={"ID": "From_action"})
        Relation.project(db=mmdb, attributes=("Flow",), svar_name="wave_enabled_flows")
        # Relation.print(db=mmdb, variable_name="wave_enabled_flows")

    def assign_waves(self):
        """
        Assign each Action in this Method Activity to a Wave of exe

        We do this by simulating the execution of Actions and propagation of Flow data.

        We start by enabling all initially available Flows and then collecting the Actions soley dependendent
        on those inputs. These Actions are assigned to the first Wave since they can execute immediately
        upon Activity invocation.

        Iteration begins with step 1: We enable all output Flows of those initial executed Actions

        Then for step 2: We find all downstream Actions yet to be executed solely dependent
        on currently enabled Flows. We assign those Actions to the next Wave, increment the wave counter
        and then repeat step 1.

        We continue until all Actions have executed and been assigned to a Wave of execution.
        """
        # Flow_Dep from_action, to_action, available,
        _logger.info("Assigning actions to waves")

        # Initialize the set of unexecuted Actions to be teh set of all Actions in this Activity
        R = f"Activity:<{self.activity_data['anum']}>, Domain:<{self.domain}>"
        Relation.restrict(db=mmdb, relation='Action', restriction=R)
        unex_actions = Relation.project(db=mmdb, attributes=("ID",), svar_name="unexecuted_actions")
        # Relation.print(db=mmdb, variable_name="unexecuted_actions")

        while unex_actions.body:
            # The loop starts with Wave 1 and then repeats the else clause until all actions have been assigned

            # Wave 1
            if self.wave_ctr == 1:
                # Set the wave_enabled_flows relation value to all initially enabled flows
                self.initially_enabled_flows()
                # Set the executable_actions relation value to all initially executable actions
                xactions = self.initially_executable_actions()

                # Initialize the completed_actions relational value to the set of executable_actions
                # We have effectively simulated their execution
                Relation.restrict(db=mmdb, relation="executable_actions", svar_name="completed_actions")
                # Relation.print(db=mmdb, variable_name="completed_actions")
                # Initialize the enabled_flows relational value to this first set of wave_enabled_flows
                # (the initially available flows)
                Relation.restrict(db=mmdb, relation="wave_enabled_flows", svar_name="enabled_flows")
                # Relation.print(db=mmdb, variable_name="enabled_flows")

                # As we move through the remaining Waves we will keep updating the set of completed_actions
                # and the set of enabled_flows until we get them all

            # Weve 2 onward
            else:
                # Proceeding from the Actions completed in the prior Wave,
                # enable the set of Flows output from those Actions
                # and make those the new wave_enabled_flows (replacing the prior set of flows)
                self.enable_action_outputs(source_action_relation="executable_actions")

                # The total set of enabled flows becomes the new wave_enabled_flows relation value added to
                # all earlier enabled flows in the enabled_flows relation value
                Relation.union(db=mmdb, relations=("enabled_flows", "wave_enabled_flows"),
                               svar_name="enabled_flows")
                # Relation.print(db=mmdb, variable_name="enabled_flows")

                # Now for the fun part:
                # For each unexecuted Action, see if its total set of required input Flows is a subset
                # of the currently enabled Flows. If so, this Action can execute and be assigned to the current Wave.

                # We use the summarize command to do the relational magic.
                # The idea is to go through the unexecuted_actions relation value and, for each Action ID
                # execute the sum_expr (summarization expression) yielding a true or false result.
                # Either the Action can or cannot execute. The unexecuted_actions relation is extended with an extra
                # column that holds this boolean result.
                # Then we'll restrict that table to find only the true results and throw away the extra column
                # in the process so that we just end up with a new set of executable_actions (replacing the previous
                # relation value of the same name).

                # Taking it step by step, here we build the sum_expr
                sum_expr = Relation.build_expr(commands=[
                    # the temproary s (summarize) relation represents the current unexecuted action
                    # We join it with the Flow Dependencies for this Activity as the To_action
                    # to obtain the set of action generated inputs it requires.
                    # (we don't care about the non-action generated initial flows since we know these are always
                    # available)
                    JoinCmd(rname1="s", rname2="fd", attrs={"ID": "To_action"}),
                    ProjectCmd(attributes=("Flow",), relation=None),
                    # And here we see if those flows are a subset of the enabled flows (true or false)
                    SetCompareCmd(rname2="enabled_flows", op=SetOp.subset, rname1=None)
                ])
                # Now we embed the sum_expr in our summarize command
                Relation.summarize(db=mmdb, relation="unexecuted_actions", per_attrs=("ID",),
                                   summaries=(
                                       SumExpr(attr=Attribute(name="Can_execute", type="boolean"), expr=sum_expr),), )
                R = f"Can_execute:<{1}>"  # True is 1, False is 0 in TclRAL and we just want the 1's
                Relation.restrict(db=mmdb, restriction=R)
                # Just take the ID attributes.  After the restrict we don't need the extra boolean attribute anymore
                xactions = Relation.project(db=mmdb, attributes=("ID",), svar_name="executable_actions")
                # Relation.print(db=mmdb, variable_name="executable_actions")
                # And here we replace the set of completed_actions with our new batch of executable_actions
                Relation.restrict(db=mmdb, relation="executable_actions", svar_name="completed_actions")

            # Having processed either the initial or subsequent waves, we do the same work
            # Add all Action IDs in the executable_actions relation into the current Wave, and increment the counter
            self.waves[self.wave_ctr] = [t['ID'] for t in xactions.body]
            self.wave_ctr += 1
            # print(f"Wave --- [{self.wave_ctr}] ---")
            # Finally, remove the latest completed actions from the set of unexecuted_actions
            unex_actions = Relation.subtract(db=mmdb, rname1="unexecuted_actions", rname2="completed_actions", svar_name="unexecuted_actions")
            # Relation.print(db=mmdb, variable_name="unexecuted_actions")
            # Rinse and repeat until all the actions are completed
        pass

    def pop_flow_dependencies(self):
        """
        For each method activity, determine the flow dependencies among its actions and populate the Flow Dependency class
        """
        # Initialize dict with key for each flow, status to be determined
        R = f"Activity:<{self.anum}>, Domain:<{self.domain}>"
        flow_r = Relation.restrict(db=mmdb, relation='Flow', restriction=R)
        flow_path = {f['ID']: {'source': set(), 'dest': set(), 'available': False} for f in flow_r.body}

        # Now proceed through each flow usage class (actions, cases, etc)
        for flow_header in flow_attrs:
            # Get all instances below the flow_header
            R = f"Activity:<{self.anum}>, Domain:<{self.domain}>"
            flow_usage_r = Relation.restrict(db=mmdb, relation=flow_header.cname, restriction=R)
            flow_usage_instances = flow_usage_r.body

            for flow_usage in flow_usage_instances:  # For each instance of this usage
                # If the header specifies an action id
                if flow_header.in_attr:
                    # Header specifies an input flow, thus a destination action
                    input_flow = flow_usage[flow_header.in_attr]
                    if input_flow in flow_path[input_flow]['dest']:
                        pass  # Dest added previously
                    flow_path[input_flow]['dest'].add(flow_usage[flow_header.id_attr])
                if flow_header.out_attr:
                    output_flow = flow_usage[flow_header.out_attr]
                    if output_flow in flow_path[output_flow]['source']:
                        pass  # Source added previously
                    flow_path[output_flow]['source'].add(flow_usage[flow_header.id_attr])

        # Mark all flows in method that are available in the first wave of execution

        # The single executing instance flow is available
        R = f"Anum:<{self.activity_data['anum']}>, Domain:<{self.domain}>"
        result = Relation.restrict(db=mmdb, relation='Method', restriction=R)
        if not result.body:
            msg = f"No executing instance populated for method: {self.domain}::{self.class_name}.{self.name}"
            _logger.error(msg)
            raise MethodXIFlowNotPopulated(msg)
        method_xi_flow = result.body[0]['Executing_instance_flow']
        flow_path[method_xi_flow]['available'] = True

        # All method parameter flows are available
        R = f"Activity:<{self.activity_data['anum']}>, Domain:<{self.domain}>"
        result = Relation.restrict(db=mmdb, relation='Parameter', restriction=R)
        for pflow in result.body:
            flow_path[pflow['Input_flow']]['available'] = True

        # All class accessor flows are available
        R = f"Activity:<{self.activity_data['anum']}>, Domain:<{self.domain}>"
        result = Relation.restrict(db=mmdb, relation='Class_Accessor', restriction=R)
        for ca_flow in result.body:
            flow_path[ca_flow['Output_flow']]['available'] = True

        for f, p in flow_path.items():
            if not (p['source'] and p['dest']):
                continue
            for source_action in p['source']:
                for dest_action in p['dest']:
                    Relvar.insert(db=mmdb, relvar='Flow_Dependency', tuples=[
                        Flow_Dependency_i(From_action=source_action, To_action=dest_action,
                                          Activity=self.activity_data['anum'], Domain=self.domain, Flow=f)
                    ])
                pass
            pass
        pass

    pass  # Method
    pass  # Populate

    def pop_xunits(self):

        _logger.info(f"Populating method execution units: {self.class_name}.{self.name}")
        # Look up signature
        R = f"Method:<{self.name}>, Class:<{self.class_name}>, Domain:<{self.domain}>"
        result = Relation.restrict(db=mmdb, relation='Method_Signature', restriction=R)
        if not result.body:
            # TODO: raise exception here
            pass
        signum = result.body[0]['SIGnum']

        # Look up xi flow
        R = f"Name:<{self.name}>, Class:<{self.class_name}>, Domain:<{self.domain}>"
        result = Relation.restrict(db=mmdb, relation='Method', restriction=R)
        if not result.body:
            # TODO: raise exception here
            pass
        xi_flow_id = result.body[0]['Executing_instance_flow']
        method_path = f"{self.domain}:{self.class_name}:{self.name}.mtd"

        aparse = self.activity_data['parse']
        activity_detail = Activity_ap(anum=self.activity_data['anum'], domain=self.domain,
                                      cname=self.class_name, sname=None, eename=None, opname=self.name,
                                      xiflow=xi_flow_id, activity_path=method_path, scrall_text=aparse[1])
        seq_flows = {}
        seq_labels = set()

        # Here we process each statement set in the Method (Activity)
        for count, xunit in enumerate(aparse[0]):  # Use count for debugging
            c = count + 1
            if type(xunit.statement_set.statement).__name__ == 'Output_Flow_a':
                # This is the statement set that returns the Method's value
                ExecutionUnit.process_synch_output(activity_data=activity_detail,
                                                   synch_output=xunit.statement_set.statement)
            else:
                # This is a statement set that does not return the Method's value
                boundary_actions = ExecutionUnit.process_method_statement_set(
                    activity_data=activity_detail, statement_set=xunit.statement_set)

            # Process any input or output tokens
            # if output_tk not in seq_flows:
            # Get a set of terminal actions
            # seq_flows[output_tk] = {'from': [terminal_actions], 'to': []}
        pass



"""
subsystem.py – Manage assignment of subsystem class and relationship numbers
"""

import logging
from xuml_populate.exceptions.mp_exceptions import CnumsExceeded


class Subsystem:
    """
    Manages the automatic assignment of unique numbers to the Subsystem's classes. In traditional Shlaer-Mellor,
    Cnums were assigned by the modeler for naming purposes just like Rnums. Here we assign and use the numbers
    internally only as a means of unique identification among all Subsystem Elements. This means that the number
    assigned to any given class may vary each time the model is populated.
    """

    def __init__(self, subsys_parse):
        """Constructor to initialize the cnum counter"""

        self._logger = logging.getLogger(__name__)

        self.name = subsys_parse['name']  # Name of the subsystem
        self.range = subsys_parse['range']  # Numbering range as a two element tuple
        self.cnum = self.range[0]  # Lowest assignable value in the range, we start counting here

    def next_cnum(self):
        """
        Assign the next available class number and throw an unrecoverable error if the numbering range is exceeded.
        """
        if self.cnum <= self.range[1]:
            self.cnum += 1
            return "C" + str(self.cnum - 1)
        else:
            self._logger.error(f"Max cnums {self.range[1]} exceeded in subsystem: {self.name}")
            raise CnumsExceeded(self.range[1])




"""
binary_association.py – Process a parsed binary association to populate the metamodel db
"""

# System
import logging

# Model Integration
from pyral.relvar import Relvar
from pyral.relation import Relation

# xUML Populate
from xuml_populate.config import mmdb
from xuml_populate.exceptions.class_exceptions import MixedTargetID, ReferenceToNonIdentifier
from xuml_populate.populate.mmclass_nt import (Association_i, Binary_Association_i, Association_Class_i,
                                               Perspective_i, Asymmetric_Perspective_i, T_Perspective_i,
                                               P_Perspective_i, Reference_i, Formalizing_Class_Role_i,
                                               Association_Reference_i, Simple_Association_Reference_i,
                                               Referring_Class_i, Attribute_Reference_i,
                                               Association_Class_Reference_i, T_Reference_i, P_Reference_i
                                               )

_logger = logging.getLogger(__name__)


class BinaryAssociation:
    """
    Populate a Binary Association
    """

    def __init__(self, tr: str, domain: str, rnum: str, record):
        """
        Populate a Binary Association

        :param tr: The name of the open transaction
        :param domain: The domain name
        :param rnum: The relationship name
        :param record: The relationship parse record
        """

        self.rnum = rnum
        self.domain = domain

        self.t_side = record.get('t_side')
        self.p_side = record.get('p_side')
        self.ref1 = record.get('ref1')
        self.ref1_source = None if not self.ref1 else self.ref1['source']
        self.ref1_target = None if not self.ref1 else self.ref1['target']
        self.ref2 = record.get('ref2')  # Supplied only for an associative binary
        self.ref2_source = None if not self.ref2 else self.ref2['source']
        self.ref2_target = None if not self.ref2 else self.ref2['target']
        self.assoc_cname = record.get('assoc_cname')
        self.assoc_mult = record.get('assoc_mult')
        self.ref1_target_id = self.targetid(self.ref1_target)
        self.ref2_target_id = None if not self.ref2_target else self.targetid(self.ref2_target)

        # Populate
        _logger.info(f"Populating Binary Association [{self.rnum}]")
        Relvar.insert(db=mmdb, tr=tr, relvar='Association', tuples=[
            Association_i(Rnum=self.rnum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr, relvar='Binary_Association', tuples=[
            Binary_Association_i(Rnum=self.rnum, Domain=domain)
        ])

        if self.assoc_cname:
            Relvar.insert(db=mmdb, tr=tr, relvar='Association_Class', tuples=[
                Association_Class_i(Rnum=self.rnum, Class=self.assoc_cname, Domain=domain, Multiplicity=self.assoc_mult)
            ])

        # Populate the T and P perspectives of an asymmetric binary association
        t_mult = self.t_side['mult'][0]
        Relvar.insert(db=mmdb, tr=tr, relvar='Perspective', tuples=[
            Perspective_i(Side='T', Rnum=self.rnum, Domain=domain,
                          Viewed_class=self.t_side['cname'], Phrase=self.t_side['phrase'],
                          Conditional=True if 'c' in self.t_side['mult'] else False,
                          Multiplicity=t_mult
                          )
        ])

        Relvar.insert(db=mmdb, tr=tr, relvar='Asymmetric_Perspective', tuples=[
            Asymmetric_Perspective_i(Side='T', Rnum=self.rnum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr, relvar='T_Perspective', tuples=[
            T_Perspective_i(Side='T', Rnum=self.rnum, Domain=domain)
        ])

        p_mult = self.p_side['mult'][0]
        Relvar.insert(db=mmdb, tr=tr, relvar='Perspective', tuples=[
            Perspective_i(Side='P', Rnum=self.rnum, Domain=domain,
                          Viewed_class=self.p_side['cname'], Phrase=self.p_side['phrase'],
                          Conditional=True if 'c' in self.p_side['mult'] else False,
                          Multiplicity=p_mult
                          )
        ])

        Relvar.insert(db=mmdb, tr=tr, relvar='Asymmetric_Perspective', tuples=[
            Asymmetric_Perspective_i(Side='P', Rnum=self.rnum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr, relvar='P_Perspective', tuples=[
            P_Perspective_i(Side='P', Rnum=self.rnum, Domain=domain)
        ])

        if not self.ref2:  # Simple binary association
            Relvar.insert(db=mmdb, tr=tr, relvar='Reference', tuples=[
                Reference_i(Ref='R',
                            From_class=self.ref1_source['class'], To_class=self.ref1_target['class'],
                            Rnum=self.rnum, Domain=domain)
            ])
            referenced_perspective = 'T' if self.ref1_target['class'] == self.t_side['cname'] else 'P'
            Relvar.insert(db=mmdb, tr=tr, relvar='Association_Reference', tuples=[
                Association_Reference_i(Ref_type='R',
                                        From_class=self.ref1_source['class'], To_class=self.ref1_target['class'],
                                        Rnum=self.rnum, Domain=domain,
                                        Perspective=referenced_perspective)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Simple_Association_Reference', tuples=[
                Simple_Association_Reference_i(Ref_type='R',
                                               From_class=self.ref1_source['class'], To_class=self.ref1_target['class'],
                                               Rnum=self.rnum, Domain=domain)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Referring_Class', tuples=[
                Referring_Class_i(Class=self.ref1_source['class'], Rnum=self.rnum, Domain=domain)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Formalizing_Class_Role', tuples=[
                Formalizing_Class_Role_i(Class=self.ref1_source['class'], Rnum=self.rnum, Domain=domain)
            ])

            # Simple Attribute Reference
            for from_attr, to_attr in zip(self.ref1_source['attrs'], self.ref1_target['attrs']):
                Relvar.insert(db=mmdb, tr=tr, relvar='Attribute_Reference', tuples=[
                    Attribute_Reference_i(From_attribute=from_attr, From_class=self.ref1_source['class'],
                                          To_attribute=to_attr, To_class=self.ref1_target['class'],
                                          Ref='R',
                                          Domain=domain, To_identifier=self.ref1_target_id, Rnum=self.rnum)
                ])
        else:  # Binary associative (with association class)
            # T Reference
            Relvar.insert(db=mmdb, tr=tr, relvar='Reference', tuples=[
                Reference_i(Ref='T',
                            From_class=self.ref1_source['class'], To_class=self.ref1_target['class'],
                            Rnum=self.rnum, Domain=domain)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Formalizing_Class_Role', tuples=[
                Formalizing_Class_Role_i(Class=self.ref1_source['class'], Rnum=self.rnum, Domain=domain)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Association_Reference', tuples=[
                Association_Reference_i(Ref_type='T',
                                        From_class=self.ref1_source['class'], To_class=self.ref1_target['class'],
                                        Rnum=self.rnum, Domain=domain,
                                        Perspective='T')
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Association_Class_Reference', tuples=[
                Association_Class_Reference_i(Ref_type='T', Association_class=self.ref1_source['class'],
                                              Participating_class=self.ref1_target['class'],
                                              Rnum=self.rnum, Domain=domain)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='T_Reference', tuples=[
                T_Reference_i(Ref_type='T', Association_class=self.ref1_source['class'],
                              Participating_class=self.ref1_target['class'],
                              Rnum=self.rnum, Domain=domain)
            ])

            # T Attribute References
            for from_attr, to_attr in zip(self.ref1_source['attrs'], self.ref1_target['attrs']):
                Relvar.insert(db=mmdb, tr=tr, relvar='Attribute_Reference', tuples=[
                    Attribute_Reference_i(From_attribute=from_attr, From_class=self.ref1_source['class'],
                                          To_attribute=to_attr, To_class=self.ref1_target['class'],
                                          Ref='T',
                                          Domain=domain, To_identifier=self.ref2_target_id, Rnum=self.rnum)
                ])

            # P Reference
            Relvar.insert(db=mmdb, tr=tr, relvar='Reference', tuples=[
                Reference_i(Ref='P',
                            From_class=self.ref2_source['class'], To_class=self.ref2_target['class'],
                            Rnum=self.rnum, Domain=domain)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Association_Reference', tuples=[
                Association_Reference_i(Ref_type='P',
                                        From_class=self.ref2_source['class'], To_class=self.ref2_target['class'],
                                        Rnum=self.rnum, Domain=domain,
                                        Perspective='P')
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Association_Class_Reference', tuples=[
                Association_Class_Reference_i(Ref_type='P', Association_class=self.ref2_source['class'],
                                              Participating_class=self.ref2_target['class'],
                                              Rnum=self.rnum, Domain=domain)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='P_Reference', tuples=[
                P_Reference_i(Ref_type='P', Association_class=self.ref2_source['class'],
                              Participating_class=self.ref2_target['class'],
                              Rnum=self.rnum, Domain=domain)
            ])

            # P Attribute References
            for from_attr, to_attr in zip(self.ref2_source['attrs'], self.ref2_target['attrs']):
                Relvar.insert(db=mmdb, tr=tr, relvar='Attribute_Reference', tuples=[
                    Attribute_Reference_i(From_attribute=from_attr, From_class=self.ref2_source['class'],
                                          To_attribute=to_attr, To_class=self.ref2_target['class'], Ref='P',
                                          Domain = domain, To_identifier = self.ref2_target_id, Rnum = self.rnum)
                ])

    def targetid(self, ref) -> int:
        # OR restriction criteria not yet supported in PyRAL, so we iterate on the attributes
        to_id = None
        for to_attr in ref['attrs']:
            R = f"Attribute:<{to_attr}>, Class:<{ref['class']}>, Domain:<{self.domain}>"
            result = Relation.restrict(db=mmdb, relation='Identifier_Attribute', restriction=R)
            if not result.body:
                _logger.exception(f"No identifier found in attribute reference on [{self.rnum}]")
                raise ReferenceToNonIdentifier
            attr_id = int(result.body[0]['Identifier'])
            if not to_id:
                to_id = attr_id
            elif to_id != attr_id:
                _logger.exception(f"Multiple identifiers referenced on [{self.rnum}]")
                raise MixedTargetID
        return to_id




""" synch_output.py - Process a Synchronous Output"""

import logging
from xuml_populate.config import mmdb
from scrall.parse.visitor import Execution_Unit_a, Seq_Statement_Set_a, Comp_Statement_Set_a
from xuml_populate.populate.statement import Statement
from xuml_populate.populate.actions.aparse_types import Activity_ap
from typing import List

class SynchOutput:
    """
    Process a Synchronous Output
    """
    _logger = logging.getLogger(__name__)

    @classmethod
    def process_method_output_flow(cls):
        pass

    @classmethod
    def process_method_statement_set(cls, activity_data: Activity_ap, statement_set) -> (List[str], List[str]):
        """
        Initiates the population of all elements derived from a set of statements in a method.

        Populates each action and returns two lists of action ids.
        The first list is each action that does not require any data input from any other action
        in the execution unit. These are initial actions since they can execute immediately.

        The second list is each action that does not provide any data input
        to any other action in the execution unit. These are terminal actions.

        :param activity_data:
        :param statement_set:
        :return: Tuple with a list of initial and terminal actions
        """
        single_statement = statement_set.statement
        block = statement_set.block
        boundary_actions = None

        # Mutually exclusive options
        if block and single_statement:
            # Parsing error, cannot have both
            raise Exception

        if single_statement:
            boundary_actions = Statement.populate(activity_data, statement_parse=single_statement)

            pass
        elif block:
            pass
        else:
            # Parsing error, neither were specified
            raise Exception


        # aid = Statement.populate()
        pass
        return boundary_actions




"""
scalar_assignment.py – Populate elements of a scalar assignment
"""

import logging
from xuml_populate.config import mmdb
from xuml_populate.populate.mmclass_nt import Labeled_Flow_i
from xuml_populate.populate.actions.extract_action import ExtractAction
from xuml_populate.populate.ns_flow import NonScalarFlow
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content, Activity_ap, Boundary_Actions
from xuml_populate.populate.actions.expressions.scalar_expr import ScalarExpr
from xuml_populate.exceptions.action_exceptions import ScalarAssignmentFlowMismatch, ScalarAssignmentfromMultipleTuples
from scrall.parse.visitor import Scalar_Assignment_a

from pyral.relvar import Relvar
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

# Transactions
tr_Migrate = "Migrate to Label"

class ScalarAssignment:
    """
    Break down a scalar assignment statement into action semantics and populate them

    """

    input_instance_flow = None  # The instance flow feeding the next component on the RHS
    input_instance_ctype = None  # The class type of the input instance flow
    domain = None
    anum = None
    activity_path = None
    scrall_text = None

    @classmethod
    def process(cls, activity_data: Activity_ap, scalar_assign_parse: Scalar_Assignment_a) -> Boundary_Actions:
        """
        Given a parsed table assignment consisting of an LHS and an RHS, populate each component action
        and return the resultant table flow

        We'll need an initial flow and we'll need to create intermediate instance flows to connect the components.
        The final output flow must be a table flow. The associated Table determines the type of the
        assignment. If the LHS spcifies an explicit Table, the resultant Table which must match.

        :param scalar_assign_parse: A parsed scalar assignment
        :param activity_data:
        """
        lhs = scalar_assign_parse.lhs
        rhs = scalar_assign_parse.rhs

        # The executing instance is by nature a single instance flow
        xi_flow = Flow_ap(fid=activity_data.xiflow, content=Content.INSTANCE, tname=activity_data.cname,
                          max_mult=MaxMult.ONE)

        bactions, scalar_flows = ScalarExpr.process(rhs=rhs, input_instance_flow=xi_flow,
                                                    activity_data=activity_data)
        scalar_flow_labels = [n for n in lhs[0].name]

        # There must be a label on the LHS for each scalar output flow
        if len(scalar_flows) != len(scalar_flow_labels):
            _logger.error(f"LHS provides {len(scalar_flow_labels)} labels, but RHS outputs {len(scalar_flows)} flows")
            raise ScalarAssignmentFlowMismatch

        # TODO: For each LHS label that explicity specifies a type, verify match with corresponding attribute in flow

        # Now we need to convert the unlabeled non scalar output_flow into one labeled scalar flow per
        # attribute in the output flow header
        # TODO: handle case where lhs is an explicit table assignment
        # Since this is a scalar flow, we need to verify that the output flow is either a single instance or tuple
        # flow with the same number of attributes as the LHS. For now let's ignore explicit typing on the LHS
        # but we need to check that later.
        for count, label in enumerate(scalar_flow_labels):
            sflow = scalar_flows[count]
            # Migrate the scalar_flow to a labeled flow
            _logger.info(f"Labeling output of scalar expression to [{lhs}]")
            Transaction.open(mmdb, tr_Migrate)
            # Delete the Unlabeled flow
            Relvar.deleteone(mmdb, tr=tr_Migrate, relvar_name="Unlabeled_Flow",
                             tid={"ID": sflow.fid, "Activity": activity_data.anum, "Domain": activity_data.domain})
            # Insert the labeled flow
            Relvar.insert(mmdb, tr=tr_Migrate, relvar='Labeled_Flow', tuples=[
                Labeled_Flow_i(ID=sflow.fid, Activity=activity_data.anum, Domain=activity_data.domain, Name=label)
            ])
            Transaction.execute(mmdb, tr_Migrate)
        return bactions

        # Create one Extract Action per attribute, label pair
        # for count, a in enumerate(attr_list):
        #     ExtractAction.populate(tuple_flow=output_flow, attr=a, target_flow_name=output_flow_labels[count],
        #                            anum=anum, domain=domain, activity_path=activity_path, scrall_text=scrall_text)



"""
extract_action.py – Populate an Extract Action instance in PyRAL
"""

import logging
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.table import Table
from typing import Set, Dict, List, Optional
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content, Activity_ap
from xuml_populate.exceptions.action_exceptions import (ProductForbidsCommonAttributes, UnjoinableHeaders,
                                                        SetOpRequiresSameHeaders)
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.mm_class import MMclass
from xuml_populate.populate.ns_flow import NonScalarFlow
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.mmclass_nt import Relational_Action_i, Extract_Action_i
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

# Transactions
tr_Extract = "Extract Action"

class ExtractAction:
    """
    Create all relations for an Extract Action
    """
    def __init__(self, tuple_flow: Flow_ap, attr: str, anum: str,
                 domain: str, activity_data: Activity_ap, label: str = None):
        """
        Populate the Extract Action

        :param tuple_flow: The input Non Scalar Flow
        :param attr: Name of attribute to extract
        :param anum: Activity number
        :param domain: Domain
        :param activity_data:
        :param label:  Name (label) of the output Scalar Flow
        """
        # Save attribute values that we will need when creating the various select subsystem
        # classes

        tuple_header = NonScalarFlow.header(ns_flow=tuple_flow, domain=domain)

        Transaction.open(db=mmdb, name=tr_Extract)
        action_id = Action.populate(tr=tr_Extract, anum=anum, domain=domain, action_type="extract")

        # Create the labeled Scalar Flow
        self.output_sflow = Flow.populate_scalar_flow(label=label, scalar_type=tuple_header[attr], anum=anum, domain=domain)

        Relvar.insert(db=mmdb, tr=tr_Extract, relvar='Relational_Action', tuples=[
            Relational_Action_i(ID=action_id, Activity=anum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Extract, relvar='Extract_Action', tuples=[
            Extract_Action_i(ID=action_id, Activity=anum, Domain=domain, Input_tuple=tuple_flow.fid,
                             Table=tuple_flow.tname, Attribute=attr, Output_scalar=self.output_sflow.fid)
        ])
        Transaction.execute(db=mmdb, name=tr_Extract)


"""
restrict_condition.py – Process a select phrase and populate a Restriction Condition
"""

import logging
from xuml_populate.config import mmdb
from xuml_populate.exceptions.action_exceptions import ActionException
from typing import Optional, Set, Dict, List
from xuml_populate.populate.attribute import Attribute
from xuml_populate.populate.actions.validation.parameter_validation import validate_param
from xuml_populate.populate.actions.table_attribute import TableAttribute
from xuml_populate.populate.actions.aparse_types import (Flow_ap, MaxMult, Content, Activity_ap, Attribute_Comparison,
                                                         Attribute_ap)
from xuml_populate.populate.actions.read_action import ReadAction
from xuml_populate.populate.actions.extract_action import ExtractAction
from xuml_populate.exceptions.action_exceptions import ComparingNonAttributeInSelection, NoInputInstanceFlow
from xuml_populate.populate.mmclass_nt import Restriction_Condition_i, Equivalence_Criterion_i, \
    Comparison_Criterion_i, Criterion_i, Table_Restriction_Condition_i
from xuml_populate.populate.flow import Flow
from pyral.relvar import Relvar
from pyral.relation import Relation
from scrall.parse.visitor import N_a, BOOL_a, Op_a, Criteria_Selection_a

_logger = logging.getLogger(__name__)

# Transactions
tr_Restrict_Cond = "Restrict Condition"


class RestrictCondition:
    """
     -> (str, List[Attribute_Comparison], Set[Flow_ap]):
    Create all relations for a Restrict Condition for either a Select or Restrict Action
        Break down criteria into a set of attribute comparisons and validate the components of a Select Action that
        must be populated into the metamodel.
         | These components are:
        * Restriction Criterian (Comparison or Ranking)
        * Scalar Flow inputs to any Comparison Criterion
        Sift through criteria to ensure that each terminal is either an attribute, input flow, or value.
        :param tr:  The select or restrict action transaction
        :param action_id:
        :param input_nsflow:
        :param activity_data:
        :param selection_parse:
        :return: Selection cardinality, attribute comparisons, and a set of scalar flows input for attribute comparison
    """
    def __init__(self, tr: str, action_id: str, input_nsflow: Flow_ap, selection_parse: Criteria_Selection_a,
                 activity_data: Activity_ap):
        """

        :param tr:
        :param action_id:
        :param input_nsflow:
        :param selection_parse:
        :param activity_data:
        """

        self.action_id = action_id
        self.anum = activity_data.anum
        self.domain = activity_data.domain
        self.activity_data = activity_data
        self.tr = tr
        self.comparison_criteria = []
        self.input_scalar_flows = set()
        self.input_nsflow = input_nsflow
        self.expression = ""
        self.criterion_ctr = 0

        criteria = selection_parse.criteria
        # Consider case where there is a single boolean value critieria such as:
        #   shaft aslevs( Stop requested )
        # The implication is that we are selecting on: Stop requested == true
        # So elaborate the parse elminating our shorthand
        self.cardinality = selection_parse.card
        if type(criteria).__name__ == 'N_a':
            self.expression = self.walk_criteria(operands=[criteria])
            pass
            # criteria = BOOL_a(op='==', operands=[criteria, N_a(name='true')])
            # Name only (no explicit operator or operand)
        else:
            self.expression = self.walk_criteria(operands=criteria.operands, operator=criteria.op)
            pass
        # Walk the parse tree and save all attributes, ops, values, and input scalar flows
        # Populate the Restriction Condition class
        Relvar.insert(db=mmdb, tr=tr, relvar='Restriction_Condition', tuples=[
            Restriction_Condition_i(Action=self.action_id, Activity=self.anum, Domain=self.domain,
                                    Expression=self.expression.strip(), Selection_cardinality=self.cardinality
                                    )
        ])
        # return cardinality, self.comparison_criteria, self.input_scalar_flows

    def walk_criteria(self, operands: List, operator: Optional[str] = None, attr: Optional[str] = None):
        """
        Recursively walk down the selection criteria parse tree validating attributes and input flows found in
        the leaf nodes. Also flatten the parse back into a language independent text representation for reference
        in the metamodel.

        :param operator:  A boolean, math or unary operator
        :param operands:  One or two operands (depending on the operator)
        :param attr:  If true, an attribute is in the process of being compared to an expression
        :return: Flattened selection expression as a string
        """
        attr_set = attr  # Has an attribute been set for this invocation?
        text = ""
        assert len(operands) <= 2
        for o in operands:
            match type(o).__name__:
                case 'IN_a':
                    # Verify that this input is defined on the enclosing Activity
                    validate_param(name=o.name, activity=self.activity_data)
                    if not attr_set:
                        # The Attribute has the same name as the Parameter
                        # We assume that attribute names are capitalized so the name doubling shorthand for
                        # params only works if this convention is followed.
                        Attribute.scalar(name=o.name.capitalize(), tname=self.input_nsflow.tname, domain=self.domain)
                        criterion_id = self.pop_xi_comparison_criterion(attr=o.name)
                    else:
                        # We know this is not an Attribute since it is a scalar flow label coming in as a Parameter
                        criterion_id = self.pop_comparison_criterion(attr=attr_set.name, op=operator, scalar_flow_label=o.name)
                    text += f" {criterion_id}"
                case 'N_a':
                    if not operator or operator in {'AND', 'OR'}:
                        if operator in {'AND', 'OR'}:
                            text += f" {operator}"
                        # This covers shorthand cases like:
                        #     ( Inservice ) -> Boolean equivalence: ( Inservice == True )
                        #     ( Held AND Blocked ) ->   same: ( Held == True AND Blocked == True )
                        # Short hand on the left and explicit equivalent longhand to the right,
                        # so the above cases leave the == True part implicit
                        # As well as name doubling shorthand:
                        #     ( Shaft ) -> Non-boolean comparison: ( Shaft == me.Shaft )
                        # The longhand form compares the value of the source instance flow attribute with the
                        # same named attribute in the xi instance. Our use of referential attributes in SM makes
                        # this sort of thing common.

                        # In all of these cases we must have an Attribute of a Class, but we verify that here
                        scalar = Attribute.scalar(name=o.name, tname=self.input_nsflow.tname, domain=self.domain)
                        # An exception is raised in the above if the Attribute is undefined

                        # Now we check the Scalar (Type) to determine whether we populate an Equivalence or
                        # Comparison Criterion
                        if scalar == 'Boolean':
                            # Populate a simple equivalence criterion
                            criterion_id = self.pop_boolean_equivalence_criterion(not_op=False, attr=o.name, value="true")
                        else:
                            # Populate a comparison
                            criterion_id = self.pop_xi_comparison_criterion(attr=o.name)
                        text += f" {criterion_id}"
                    elif not attr_set:
                        # We know that the operator is a comparison op (==, >=, etc) which means that
                        # the first operand is an Attribute and the second is a scalar expression
                        # TODO: ignore NOT operator for now
                        # And that since the attribute hasn't been set yet, this must be the left side of the
                        # comparison and, hence, the attribute
                        # We verify this:
                        scalar = Attribute.scalar(name=o.name, tname=self.input_nsflow.tname, domain=self.domain)
                        # The criterion is populated when the second operand is processed, so all we need to do
                        # now is to remember the Attribute name
                        attr_set = Attribute_ap(o.name, scalar)
                    else:
                        # The scalar expression on the right side of the comparison must be a scalar flow
                        # an enum, or a boolean value
                        if o.name.startswith('_'):
                            # This name is an enum value
                            # TODO: Validate the enum value as a member of the Attribute's Scalar
                            criterion_id = self.pop_equivalence_criterion(attr=attr_set.name, op=operator, value=o.name,
                                                          scalar=attr_set.scalar)
                        elif (n := o.name.lower()) in {'true', 'false'}:
                            # This name is a boolean value
                            criterion_id = self.pop_boolean_equivalence_criterion(not_op=False, attr=attr_set, value=n)
                        else:
                            # It must be the name of a scalar flow that should have been set with some value
                            criterion_id = self.pop_comparison_criterion(scalar_flow_label=o.name, attr=attr_set.name, op=operator)
                        text += f" {criterion_id}"

                case 'BOOL_a':
                    criterion_id = self.walk_criteria(operands=o.operands, operator=o.op, attr=attr_set)
                    text += f" {criterion_id}"
                case 'MATH_a':
                    criterion_id = self.walk_criteria(operands=o.operands, operator=o.op, attr=attr_set)
                    text += f" {criterion_id}"
                case 'UNARY_a':
                    pass # TODO: tbd
                case 'INST_PROJ_a':
                    match type(o.iset).__name__:
                        case 'N_a':
                            if o.projection:
                                # This must be a Non Scalar Flow
                                # If it is an Instance Flow, an attribute will be read with a Read Action
                                # Otherwise, a Tuple Flow will have a value extracted with an Extract Action

                                sflow = None  # This is the scalar flow result of the projection/extraction
                                ns_flow = Flow.find_labeled_ns_flow(name=o.iset.name, anum=self.anum, domain=self.domain)
                                if not ns_flow:
                                    raise ActionException
                                if ns_flow.content == Content.INSTANCE:
                                    # TODO: Fill out the read action case
                                    ReadAction.populate()
                                elif ns_flow.content == Content.RELATION:
                                    if len(o.projection.attrs) != 1:
                                        # For attribute comparison, there can only be one extracted attribute
                                        raise ActionException
                                    attr_to_extract = o.projection.attrs[0].name
                                    extract_action = ExtractAction(
                                        tuple_flow=ns_flow, attr=attr_to_extract, anum=self.anum,
                                        domain=self.domain, activity_data=self.activity_data
                                    )  # Select Action transaction is open
                                    sflow = extract_action.output_sflow
                                # Now populate a comparison criterion
                                criterion_id = self.pop_comparison_criterion(attr=o.projection.attrs[0].name, scalar_flow=sflow, op=operator)
                                text += f" {criterion_id}"
                            else:
                                # This must be a Scalar Flow
                                # TODO: check need for mmdb param
                                sflow = Flow.find_labeled_scalar_flow(name=o.iset.name, anum=self.anum,
                                                                      domain=self.domain)
                                if not sflow:
                                    raise ActionException
                            pass
                        case 'IN_a':
                            pass
                        case 'INST_a':
                            pass
                            # TODO: create what's needed when we hit this case (previously order expr)
                        case _:
                            raise ActionException
                    # TODO: Now process the projection
                    pass
                case _:
                    raise Exception
        return text

    def pop_xi_comparison_criterion(self, attr: str) -> int:
        """
        Let's say that we are performing a select/restrict on some Class and comparing the value of some
        attribute named X on that Class. If the executing instance (xi) also has an attribute named X,
        we can read that value and supply it in the comparison criterion.

        Here we populate a Read Action on the executing instance (xi) to read the value of the matching
        Attribute.

        :param attr: Name of some compared Attribute that matches an Attribute of the executing instance
        """
        read_iflow = Flow_ap(fid=self.activity_data.xiflow, tname=self.activity_data.cname, content=Content.INSTANCE,
                             max_mult=MaxMult.ONE)
        _, read_flows = ReadAction.populate(input_single_instance_flow=read_iflow,
                                            attrs=(attr,),
                                            anum=self.anum, domain=self.domain)
        assert len(read_flows) == 1
        # Since we are reading a single attribute, assume only one output flow
        return self.pop_comparison_criterion(attr=attr, op='==', scalar_flow=read_flows[0])

    def pop_comparison_criterion(self, attr: str, op: str, scalar_flow_label: Optional[str] = None,
                                 scalar_flow: Optional[Flow_ap] = None) -> int:
        """

        :param attr:
        :param op:
        :param scalar_flow_label:
        :param scalar_flow:
        :return: criterion id
        """
        if not scalar_flow:
            if not scalar_flow_label:
                raise ActionException
            sflow = Flow.find_labeled_scalar_flow(name=scalar_flow_label, anum=self.anum, domain=self.domain)
        else:
            sflow = scalar_flow
        self.input_scalar_flows.add(sflow)
        if not sflow:
            raise ActionException  # TODO: Make specific
        criterion_id = self.pop_criterion(attr)
        Relvar.insert(db=mmdb, tr=self.tr, relvar='Comparison_Criterion', tuples=[
            Comparison_Criterion_i(ID=criterion_id, Action=self.action_id, Activity=self.anum, Attribute=attr,
                                   Comparison=op, Value=sflow.fid, Domain=self.domain)
        ])
        self.comparison_criteria.append(Attribute_Comparison(attr, op))
        return criterion_id

    def pop_criterion(self, attr: str) -> int:
        """

        :param attr:
        :return: criterion id
        """
        self.criterion_ctr += 1
        criterion_id = self.criterion_ctr
        Relvar.insert(db=mmdb, tr=self.tr, relvar='Criterion', tuples=[
            Criterion_i(ID=criterion_id, Action=self.action_id, Activity=self.anum, Attribute=attr,
                        Non_scalar_type=self.input_nsflow.tname, Domain=self.domain)
        ])
        return criterion_id

    def pop_equivalence_criterion(self, attr: str, op: str, value: str, scalar: str) -> int:
        """
        Populates either a boolean or enum equivalence

        :param attr: Attribute name
        :param op: Either eq or ne (== !=)
        :param value: Enum value or true
        :param scalar: Scalar name
        """
        # Populate the Restriction Criterion superclass
        criterion_id = self.pop_criterion(attr=attr)
        # Populate the Equivalence Criterion
        Relvar.insert(db=mmdb, tr=self.tr, relvar='Equivalence_Criterion', tuples=[
            Equivalence_Criterion_i(ID=criterion_id, Action=self.action_id, Activity=self.anum,
                                    Attribute=attr, Domain=self.domain, Operation=op,
                                    Value=value, Scalar=scalar)
        ])
        return criterion_id

    def pop_boolean_equivalence_criterion(self, not_op: bool, attr: str, value: str) -> int:
        """
        An Equivalence Criterion is populated when a boolean value is compared
        against an Attribute typed Boolean

        :param not_op: True if attribute preceded by NOT operator in expression
        :param attr: Attribute name
        :param value:  Attribute is compared to either "true" / "false"
        :return: criterion id
        """
        return self.pop_equivalence_criterion(attr=attr, op="ne" if not_op else "eq", value=value, scalar="Boolean")




"""
statement.py – Populate all actions in a Scrall statement
"""

import logging
from xuml_populate.populate.actions.instance_assignment import InstanceAssignment
from xuml_populate.populate.actions.table_assignment import TableAssignment
from xuml_populate.populate.actions.scalar_assignment import ScalarAssignment
from xuml_populate.populate.actions.switch_statement import SwitchStatement
from xuml_populate.populate.actions.aparse_types import Activity_ap, Boundary_Actions, Labeled_Flow
from collections import namedtuple
from typing import Set

_logger = logging.getLogger(__name__)

class Statement:
    """
    Create all relations for a Statement
    """
    next_action_id = {}
    activity_type = None  # enum: state, ee, method
    state = None  # state name
    operation = None  # operation name
    method = None  # method name
    cname = None
    xi_flow_id = None

    @classmethod
    def populate(cls, activity_data: Activity_ap, statement_parse: namedtuple,
                 case_name: str = '', case_outputs: Set[Labeled_Flow] = None) -> Boundary_Actions:
        """
        Populate a Statement

        :param activity_data:
        :param statement_parse:  A single parsed statement for us to populate
        :param case_name:  Values matched by case concatenated into a string
        :param case_outputs: Each output labeled data flow in the case, if any
        :return:
        """
        statement_type = type(statement_parse).__name__
        # For now we'll just switch on the action_group name and later wrap all this up
        # into a dictionary of functions of some sort
        match statement_type:
            case 'Inst_Assignment_a':
                boundary_actions = InstanceAssignment.process(activity_data=activity_data,
                                                              inst_assign=statement_parse,
                                                              case_name=case_name,
                                                              case_outputs=case_outputs,
                                                              )
                pass
            case 'Table_Assignment_a':
                boundary_actions = TableAssignment.process(activity_data=activity_data,
                                                           table_assign_parse=statement_parse,
                                                           case_name=case_name,
                                                           case_outputs=case_outputs)
                pass
            case 'Scalar_Assignment_a':
                boundary_actions = ScalarAssignment.process(activity_data=activity_data,
                                                            scalar_assign_parse=statement_parse,
                                                            # case_outputs=case_outputs,
                                                            # case_prefix=case_prefix
                                                            )
                pass
            case 'Switch_a':
                boundary_actions = SwitchStatement.populate(activity_data=activity_data,
                                                            sw_parse=statement_parse)
            case _:
                boundary_actions = None
                # print()

        return boundary_actions



"""
mm_class.py – Process parsed class to populate the metamodel db
"""

import logging
from xuml_populate.config import mmdb
from pyral.transaction import Transaction
from pyral.relvar import Relvar
from xuml_populate.populate.element import Element
from xuml_populate.populate.attribute import Attribute
from xuml_populate.populate.mm_type import MMtype
from xuml_populate.populate.mmclass_nt import Class_i, Alias_i

from pyral.relation import Relation

from typing import Dict

_logger = logging.getLogger(__name__)

# Transactions
_tr_Class = "Class"


class MMclass:
    """
    Populate all relevant Class relvars
    """
    record = None
    name = None
    alias = None
    cnum = None
    identifiers = None
    attributes = None
    methods = None
    ee = None
    ee_ops = None

    @classmethod
    def header(cls, cname: str, domain: str) -> Dict[str, str]:
        """
        Returns the header for a table reprsentation of this class

        :param cname:   Class name
        :param domain:  Domain name
        :return: The header as a dictionary of attr;type key value pairs
        """
        R = f"Class:<{cname}>, Domain:<{domain}>"
        attrs = Relation.restrict(mmdb, relation='Attribute', restriction=R)
        h = {a['Name']: a['Scalar'] for a in attrs.body}
        return h

    @classmethod
    def exists(cls, cname: str, domain: str) -> bool:
        """

        :param cname:  Name of the class
        :param domain: Its domain name
        :return: True if the class has been populated into this domain
        """
        R = f"Name:<{cname}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, relation='Class', restriction=R)
        return bool(result.body)



    @classmethod
    def populate(cls, domain: str, subsystem, record):
        """
        Populate a metamodel Class relation

        :param domain:
        :param subsystem:
        :param record:
        :return:
        """

        cls.record = record
        cls.name = record['name']
        cls.attributes = record['attributes']
        cls.alias = record.get('alias')  # Optional
        cls.methods = record.get('methods')
        cls.ee = record.get('ee')
        cls.ee_ops = record.get('ee_ops')

        # Get the next cnum
        cls.cnum = subsystem.next_cnum()
        #
        # Populate class
        _logger.info(f"Populating class [{cls.name}]")
        _logger.info("Transaction open: Populate class")
        Transaction.open(mmdb, "Class")  # Class, Class Type and Attributes

        # Populate the corresponding Type superclass
        MMtype.populate_class(tr=_tr_Class, cname=cls.name, domain=domain)

        Element.populate_labeled_subys_element(tr=_tr_Class, label=cls.cnum,
                                               subsystem=subsystem.name, domain=domain)
        Relvar.insert(mmdb, tr=_tr_Class, relvar='Class', tuples=[
            Class_i(Name=cls.name, Cnum=cls.cnum, Domain=domain)
        ])
        if cls.alias:
            Relvar.insert(mmdb, tr=_tr_Class, relvar='Alias', tuples=[
                Alias_i(Name=cls.name, Class=cls.name, Domain=domain)
            ])

        # Populate the attributes
        cls.identifiers = set()  # For each newly created class we clear the id set
        for a in cls.record['attributes']:
            Attribute.populate(tr=_tr_Class, domain=domain, cname=cls.name,
                               class_identifiers=cls.identifiers, record=a)

        Transaction.execute(mmdb, _tr_Class)  # Class, Class Type, and Attributes
        _logger.info("Transaction closed: Populate class")



"""
method.py – Process parsed method to populate the metamodel db
"""
# System
import logging

# Model Integration
from pyral.transaction import Transaction
from pyral.relvar import Relvar
from pyral.relation import Relation  # For debugging

# xUML Populate
from xuml_populate.config import mmdb
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.signature import Signature
from xuml_populate.populate.activity import Activity
from xuml_populate.populate.mm_type import MMtype
from xuml_populate.populate.mmclass_nt import (Method_Signature_i, Method_i, Parameter_i, Synchronous_Output_i)

_logger = logging.getLogger(__name__)

# Transactions
tr_Method = "Method"
tr_Parameter = "Parameter"
tr_OutputFlow = "OutputFlow"


class Method:
    """
    Populate all relevant Method relvars
    """
    def __init__(self, domain: str, subsys: str, m_parse, parse_actions: bool):
        """
        Populate a method

        :param domain: The name of the domain
        :param subsys: The name of the subsystem
        :param m_parse: The parsed content of the method
        :return:
        """
        subsys_method_path = None
        me_flow = None  # Executing instance flow

        class_name = m_parse.class_name

        Transaction.open(db=mmdb, name=tr_Method)
        _logger.info("Transaction open: Populating method")

        # Create the signature
        signum = Signature.populate(tr=tr_Method, subsys=subsys, domain=domain)
        Relvar.insert(db=mmdb, tr=tr_Method, relvar='Method_Signature', tuples=[
            Method_Signature_i(SIGnum=signum, Method=m_parse.method, Class=class_name, Domain=domain)
        ])

        # Populate the method
        anum = Activity.populate_method(tr=tr_Method, action_text=m_parse.activity,
                                        cname=class_name,
                                        method=m_parse.method,
                                        subsys=subsys, domain=domain, parse_actions=parse_actions)

        # Populate the executing instance (me) flow
        self.me_flow = Flow.populate_instance_flow(cname=class_name, anum=anum, domain=domain,
                                                   label='me', single=True, activity_tr=tr_Method)
        _logger.info(f"INSERT Instance Flow (method me): [{domain}:{class_name}:{m_parse.method}:"
                     f"{self.me_flow.fid}]")
        Relvar.insert(db=mmdb, tr=tr_Method, relvar='Method', tuples=[
            Method_i(Anum=anum, Name=m_parse.method, Class=class_name, Domain=domain,
                     Executing_instance_flow=self.me_flow.fid)
        ])

        Transaction.execute(db=mmdb, name=tr_Method)  # Populate empty method
        _logger.info("Transaction closed: Populating method")

        # Add input flows (parameters)
        for p in m_parse.flows_in:
            # Populate the Parameter's type if it hasn't already been populated
            MMtype.populate_unknown(name=p['type'], domain=domain)

            _logger.info("Transaction open: Populating method parameter")
            Transaction.open(db=mmdb, name=tr_Parameter)

            input_fid = Flow.populate_data_flow_by_type(mm_type=p['type'], anum=anum,
                                                        domain=domain, label=p['name'],activity_tr=tr_Parameter).fid

            _logger.info(f"INSERT Scalar Flow (method input): ["
                         f"{domain}:{class_name}:{m_parse.method}:^{p['name']}:{input_fid}]")
            Relvar.insert(db=mmdb, tr=tr_Parameter, relvar='Parameter', tuples=[
                Parameter_i(Name=p['name'], Signature=signum, Domain=domain,
                            Input_flow=input_fid, Activity=anum, Type=p['type'])
            ])
            Transaction.execute(db=mmdb, name=tr_Parameter)  # Method parameter
            _logger.info("Transaction closed: Populating parameter")

        # Output flow (created by output flow action when it is populated)



""" xunit.py - Process a Scrall Execution Unit"""

# System
import logging
from typing import List

# Model Integration
from scrall.parse.visitor import Output_Flow_a
from pyral.relvar import Relvar

# Xuml Populate
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.aparse_types import Flow_ap, Content, MaxMult
from xuml_populate.populate.statement import Statement
from xuml_populate.populate.actions.aparse_types import Activity_ap
from xuml_populate.populate.actions.expressions.instance_set import InstanceSet
from xuml_populate.populate.actions.expressions.scalar_expr import ScalarExpr
from xuml_populate.populate.mmclass_nt import Synchronous_Output_i
from xuml_populate.exceptions.action_exceptions import *

tr_OutputFlow = "OutputFlow"

_logger = logging.getLogger(__name__)

class ExecutionUnit:
    """
    Process an Execution Unit
    """

    @classmethod
    def process(cls):
        pass

    @classmethod
    def process_statement_set(cls) -> (List[str], List[str]):
        pass

    @classmethod
    def process_synch_output(cls, activity_data: Activity_ap, synch_output: Output_Flow_a):
        """

        :param activity_data:
        :param synch_output:  Output flow execution unit parse
        :return:
        """
        cls.activity_data = activity_data
        xi_instance_flow = Flow_ap(fid=activity_data.xiflow, content=Content.INSTANCE, tname=activity_data.cname,
                                   max_mult=MaxMult.ONE)
        match type(synch_output.output).__name__:
            case 'INST_a':
                _, _, output_flow = InstanceSet.process(input_instance_flow=xi_instance_flow,
                                                        iset_components=synch_output.output.components,
                                                        activity_data=activity_data)
                pass
            case 'INST_PROJ_a':
                _, _, output_flow = InstanceSet.process(input_instance_flow=xi_instance_flow,
                                                        iset_components=synch_output.output.iset.components,
                                                        activity_data=activity_data)
            case _:
                # Unexpected or unimplemented synch output case
                msg = f"No case for synch output exec unit type: [{type(synch_output.output).__name__}]"
                _logger.error(msg)
                raise UndefinedSynchOutputExecutionUnit(msg)
        # b, f = ScalarExpr.process(mmdb, rhs=synch_output.output, input_instance_flow=xi_instance_flow,
        #                           activity_data=activity_data)

        # Populate the output flow (no transaction required)
        Relvar.insert(db=mmdb, relvar='Synchronous_Output', tuples=[
            Synchronous_Output_i(Anum=activity_data.anum, Domain=activity_data.domain,
                                 Output_flow=output_flow.fid, Type=output_flow.tname)
        ])
        _logger.info(f"INSERT Synchronous operation output flow): ["
                     f"{activity_data.domain}:{activity_data.cname}:{activity_data.opname}:^{output_flow.fid}]")

    @classmethod
    def process_state_statement_set(cls):
        pass

    @classmethod
    def process_operation_output_flow(cls):
        pass

    @classmethod
    def process_operation_statement_set(cls):
        pass

    @classmethod
    def process_method_output_flow(cls):
        pass

    @classmethod
    def process_method_statement_set(cls, activity_data: Activity_ap, statement_set) -> (
            List[str], List[str]):
        """
        Initiates the population of all elements derived from a set of statements in a method.

        Populates each action and returns two lists of action ids.
        The first list is each action that does not require any data input from any other action
        in the execution unit. These are initial actions since they can execute immediately.

        The second list is each action that does not provide any data input
        to any other action in the execution unit. These are terminal actions.

        :param activity_data:  Info about the activity and its unparsed text. Useful for providing helpful error msgs
        :param statement_set:  The statement set we are populating
        :return: Tuple with a list of initial and terminal actions
        """
        single_statement = statement_set.statement
        block = statement_set.block
        boundary_actions = None

        # Mutually exclusive options
        if block and single_statement:
            # Parsing error, cannot have both
            raise Exception

        if single_statement:
            boundary_actions = Statement.populate(activity_data=activity_data, statement_parse=single_statement)

            pass
        elif block:
            pass  # TODO: Implement this when we have an example
        else:
            # Parsing error, neither were specified
            raise Exception

        # aid = Statement.populate()
        pass
        return boundary_actions



""" instance_set.py """

import logging
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.traverse_action import TraverseAction
from xuml_populate.populate.actions.expressions.class_accessor import ClassAccessor
from xuml_populate.populate.mm_class import MMclass
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.actions.select_action import SelectAction
from xuml_populate.populate.actions.restrict_action import RestrictAction
from xuml_populate.populate.actions.rank_restrict_action import RankRestrictAction
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content, Activity_ap
from pyral.relation import Relation
from pyral.transaction import Transaction
from xuml_populate.exceptions.action_exceptions import NoClassOrInstanceFlowForInstanceSetName, \
    SelectionOnScalarFlow

_logger = logging.getLogger(__name__)


class InstanceSet:
    """
    An instance set is a Scrall grammar element representing a chain of one or more components. Each component is a
    parse element that populates as one or more Actions or Flows. Each Action outputs an instance flow which is fed
    into the next Action until a final Instance Flow output is generated.

    Our goal is to process each component and populate the associated Actions and Flows. Along the way, we verify
    that the requested elements are consistent with the underlying class model and any Flows previously defined
    in the same Activity. An exception is thrown if there is any inconsistency.
    """

    component_flow = None  # Output flow from the most recently populated component
    initial_action = None  # The first action in the chain
    final_action = None  # The last action in the chain

    @classmethod
    def process(cls, input_instance_flow: Flow_ap, iset_components,
                activity_data: Activity_ap) -> ( str, str, Flow_ap):
        """
        Populate any Actions or Flows corresponding to a sequence of instance set components.
        Return the boundary actions and the resultant output Instance Flow

        :param input_instance_flow: This Flow provides input at the beginning of the chain
        :param iset_components: The components in the instance set
        :param activity_data:  General information about the enclosing anum, anum, domain, etc
        :return: The initial action id, the final action id, and the output instance flow
        """
        cls.initial_action = None  # Clear out whatever old values might remain from a prior instance set
        cls.final_action = None
        cls.component_flow = input_instance_flow  # This will be input to the first component
        domain = activity_data.domain
        anum = activity_data.anum

        first_action = True  # We use this to recognize the initial action
        for count, comp in enumerate(iset_components):
            # We use the count to recognize the final action
            match type(comp).__name__:
                case 'PATH_a':
                    # Path component
                    # Process the path to create the traverse action and obtain the resultant output instance flow
                    traverse_action = TraverseAction(input_instance_flow=cls.component_flow, path=comp,
                                                     activity_data=activity_data)
                    aid = traverse_action.action_id
                    cls.component_flow = traverse_action.output_flow

                    # Data flow to/from actions within the instance_set
                    if first_action:
                        # For the first component, there can be dflow input from another action
                        cls.initial_action = aid
                        first_action = False  # The first action has been encountered and recognized as initial
                    if count == len(iset_components) - 1:
                        # For the last component, there can be no dflow output to another action
                        cls.final_action = aid
                case 'N_a':
                    # Name component
                    # Is it a class name?  If so, we'll need a Class Accessor populated if we don't have one already
                    class_flow = ClassAccessor.populate(name=comp.name, anum=anum, domain=domain)
                    if class_flow:
                        # We have a Class Accessor either previously or just now populated
                        # Set its output flow to the current component output
                        cls.component_flow = Flow_ap(fid=class_flow, content=Content.INSTANCE,
                                                     tname=comp.name, max_mult=MaxMult.MANY)
                    else:
                        # Is it a Non Scalar Flow?
                        ns_flow = Flow.find_labeled_ns_flow(name=comp.name, anum=anum, domain=domain)
                        if ns_flow:
                            cls.component_flow = ns_flow
                        else:
                            # Either there is no corresponding flow or it is a Scalar Flow
                            raise NoClassOrInstanceFlowForInstanceSetName(path=activity_data.activity_path,
                                                                          text=activity_data.scrall_text,
                                                                          x=iset_components.X)
                case 'Criteria_Selection_a':
                    # Process to populate a select action, the output type does not change
                    # since we are selecting on a known class
                    if cls.component_flow.content == Content.INSTANCE:
                        sa = SelectAction(
                            input_instance_flow=cls.component_flow, selection_parse=comp,
                            activity_data=activity_data
                        )
                        aid = sa.action_id
                        cls.component_flow = sa.output_instance_flow
                        sflows = sa.sflows
                    elif cls.component_flow.content == Content.RELATION:
                        ra = RestrictAction(
                            input_relation_flow=cls.component_flow, selection_parse=comp,
                            activity_data=activity_data
                        )
                        aid = ra.action_id
                        cls.component_flow = ra.output_relation_flow
                        sflows = ra.sflows
                    else:
                        pass  # TODO: Exception (can't be scalar)
                    # Data flow to/from actions within the instance_set
                    if first_action:
                        # For the first component, there can be dflow input from another action
                        cls.initial_action = aid
                        first_action = False  # The first action has been encountered and recognized as initial
                    if count == len(iset_components) - 1:
                        # For the last component, there can be no dflow output to another action
                        cls.final_action = aid
                case 'Rank_Selection_a':
                    ranksel = RankRestrictAction(input_relation_flow=cls.component_flow, selection_parse=comp,
                                                 activity_data = activity_data)
                    aid = ranksel.action_id
                    cls.component_flow = ranksel.output_relation_flow
                    # Data flow to/from actions within the instance_set
                    if first_action:
                        # For the first component, there can be dflow input from another action
                        cls.initial_action = aid
                        first_action = False  # The first action has been encountered and recognized as initial
                    if count == len(iset_components) - 1:
                        # For the last component, there can be no dflow output to another action
                        cls.final_action = aid
                case _:
                    raise Exception
        return cls.initial_action, cls.final_action, cls.component_flow



"""
tree.py – Functions that process tree elements
"""

from typing import List

def extract(tree: List) -> (List[str], bool):
    """
    A tree is represented as a list of strings and sublists.
    A single invocation grabs all strings up to the first sublist, and then invokes itself on that sublist.
    Eventually, a sublist is encountered that is simply a list of strings terminating the recursion.
    As each invocation returns, any visited sublist is deleted.

    Recursive execution of this function effects a left handed search through the tree
    to extract a flat branch as a list of strings only and to prune that branch from the source tree.

    Therefore, this function has a side effect of pruning the provided tree

    :param tree: A list of strings and sublists. It may be modified by this function.
    :return: The pruned branch and whether or not this invocation of extract encountered a terminal (string only list).
    The result is useful internally to terminate recursion, but of no value to the outside caller
    """
    pop = []  # population
    # If there are no sublists in the node, it is a terminal node and we can just return its contents
    if not any(isinstance(n, list) for n in tree):
        # print("Terminal node")
        return tree, True
    # Step through each node
    for i, n in enumerate(tree):
        # If the node is a string, just append it to the population
        if type(n) is str:
            pop.append(n)
        else:  # Otherwise it is nested and we extract a branch from it
            branch = extract(n)
            pop = pop + branch[0]  # Add the extracted branch to the population
            if branch[1]:  # the extracted branch was a terminal node (no deeper nesting)
                del tree[i]  # We delete it from the tree so we don't process it again
                # If the remaining tree is a parent with only one child, we can transform it into
                # a flat branch. This means we have ['n0', ['n1']] and we can flatten that into
                # [ 'n0', 'n1' ]. This keeps us from ending up with a childless parent.
                if len(tree) == 2 and len(tree[1]) == 1:
                    tree[1] = tree[1][0]
                # print()
            break  # Once we process a nested node, we don't go any further
    # print()
    return pop, False  # This is not a terminal node


pattern = ['A', 'B', ['C', ['D', 'E'], ['F']], ['G', 'H']]

lineages = []
if not any(isinstance(n, list) for n in pattern):
    lineages = pattern
else:
    while len(pattern) > 0 and any(isinstance(n, list) for n in pattern):
        lineages.append(extract(pattern)[0])
# print()


"""
rank_restrict_action.py – Populate a Rank Restrict Action instance in PyRAL
"""

import logging
from typing import Set
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.aparse_types import Flow_ap, Activity_ap
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.actions.expressions.restriction_condition import RestrictCondition
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.mmclass_nt import Relational_Action_i, Table_Action_i, Rank_Restrict_Action_i
from pyral.relvar import Relvar
from pyral.relation import Relation  # Here for debugging
from pyral.transaction import Transaction
from scrall.parse.visitor import Criteria_Selection_a, Rank_Selection_a

_logger = logging.getLogger(__name__)

# Transactions
tr_Rank_Restrict_Action = "Rank Restrict Action"

class RankRestrictAction:
    """
    Create all relations for a Restrict Action
    """

    def __init__(self, input_relation_flow: Flow_ap, selection_parse: Rank_Selection_a,
                 activity_data: Activity_ap):
        """
        Populate the Rank Restrict Action
         -> (str, Flow_ap, Set[Flow_ap])
        :param input_relation_flow: The source table flow into this restriction
        :param selection_parse:  The parsed Scrall select action group
        :param activity_data:
        :return: The select action id, the output flow, and any scalar flows input for attribute comparison
        """
        # Save attribute values that we will need when creating the various select subsystem
        # classes
        domain = activity_data.domain
        anum = activity_data.anum

        # Populate the Action superclass instance and obtain its action_id
        Transaction.open(db=mmdb, name=tr_Rank_Restrict_Action)
        self.action_id = Action.populate(tr=tr_Rank_Restrict_Action, anum=anum, domain=domain, action_type="rank restrict")

        # Populate the output Table Flow using same Table as input flow
        # TODO: This is a tuple flow if the cardinality is one
        self.output_relation_flow = Flow.populate_relation_flow_by_reference(
            ref_flow=input_relation_flow, anum=anum, domain=domain, tuple_flow=selection_parse.card == 'ONE')

        Relvar.insert(db=mmdb, tr=tr_Rank_Restrict_Action, relvar='Relational_Action', tuples=[
            Relational_Action_i(ID=self.action_id, Activity=anum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Rank_Restrict_Action, relvar='Table_Action', tuples=[
            Table_Action_i(ID=self.action_id, Activity=anum, Domain=domain,
                           Input_a_flow=input_relation_flow.fid, Output_flow=self.output_relation_flow.fid)
        ])
        Relvar.insert(db=mmdb, tr=tr_Rank_Restrict_Action, relvar='Rank_Restrict_Action', tuples=[
            Rank_Restrict_Action_i(ID=self.action_id, Activity=anum, Domain=domain,
                                   Attribute=selection_parse.attr, Non_scalar_type=input_relation_flow.tname,
                                   Selection_cardinality=selection_parse.card, Extent=selection_parse.rankr)
        ])
        # We now have a transaction with all select-action instances, enter into the metamodel db
        Transaction.execute(db=mmdb, name=tr_Rank_Restrict_Action)  # Restrict Action


"""
mp_exceptions.py – Model parser specific exceptions
"""

# Every error should have the same format
# with a standard prefix and postfix defined here
pre = "\nxUML model parser: ["
post = "]"


class MPException(Exception):
    pass

class MDPopulationException(MPException):
    pass

class BadStateModelName(MPException):
    pass

class MultipleDomainsException(MPException):
    def __str__(self):
        return f'{pre}Only one domain may be loaded at a time.{post}'

class UnknownRelationshipType(MDPopulationException):
    def __str__(self):
        return f'{pre}Relationship is not a Generalization, Ordinal or Assocation.{post}'

class CnumsExceeded(MDPopulationException):
    def __init__(self, maxcnum):
        self.maxcnum = maxcnum

    def __str__(self):
        return f'{pre}Exceeded maximum cnum {self.maxcnum} for subsystem. Adjust number range.{post}'

class MismatchedStateSignature(MDPopulationException):
    def __init__(self, event, state):
        self.event = event
        self.state = event

    def __str__(self):
        return f'{pre}Event spec <{self.event}> on transition into state [{self.state}] has a different signature.{post}'

class LnumsExceeded(MDPopulationException):
    def __init__(self, maxlnum):
        self.maxlnum = maxlnum

    def __str__(self):
        return f'{pre}Exceeded maximum lnum (lineage number) {self.maxlnum} for subsystem. Adjust number range.{post}'

class LessThanTwoSubclassesInGeneralization(MDPopulationException):
    def __init__(self, rnum):
        self.rnum = rnum

    def __str__(self):
        return f'{pre}Generalization [{rnum}] requires at least two subclasses to form a minimal paritition.{post}'

class MPIOException(MPException):
    pass

class MPDBException(MPException):
    pass


class MPUserInputException(MPException):
    pass

class ModelParseError(MPUserInputException):
    def __init__(self, model_file, e):
        self.model_file = model_file
        self.e = e

    def __str__(self):
        return f'{pre}Parse error in model "{self.model_file}"\n\t{self.e}"{post}'

class ModelInputFileOpen(MPIOException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return f'{pre}Parser cannot open this model input file: "{self.path}"{post}'

class ModelInputFileEmpty(MPIOException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return f'{pre}For some reason, nothing was read from the model input file: "{self.path}"{post}'

class ModelGrammarFileOpen(MPIOException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return f'{pre}Parser cannot open this model grammar file: "{self.path}"{post}'


"""
operation.py – Convert parsed operation to a relation
"""

import logging
from xuml_populate.config import mmdb
from pyral.transaction import Transaction
from pyral.relvar import Relvar
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.signature import Signature
from xuml_populate.populate.activity import Activity
from xuml_populate.populate.mm_type import MMtype
from xuml_populate.populate.mmclass_nt import Operation_Signature_i, Operation_i, Parameter_i,\
    Asynchronous_Operation_i, Synchronous_Operation_i, Synchronous_Output_i

_logger = logging.getLogger(__name__)

tr_Op = "Operation"
tr_EE = "EE"
tr_Parameter = "Parameter"
tr_Output = "Synchronous Output"

class Operation:
    """
    Create a operation relation
    """

    @classmethod
    def populate(cls, tr: str, subsys: str, domain: str, parsed_op) -> str:
        """
        Populate a single operation for an EE

        (This is meant to be invoked for each operation of the EE)

        :param tr: Either "EE" or "Operation" as the name of the transaction
        :param subsys: The subsystem name
        :param domain: The domain name
        :param parsed_op:  The parsed operation
        :return: Change the transaction from "EE" to "Operation" for subsequent invocations
        """
        # The first invocation for an EE will be tr_EE in which case we know that the EE transaction is
        # still pending and we want to use that transaction to populate its first required Operation
        # Upon completion we return "Operation" as the ongoing transaction until the next EE comes along

        if tr != tr_EE:
            # On the first op we already have an open transaction for the EE population
            # For each subsequent op, a new transaction must be opened
            logging.info("Transaction open: Operation")
            Transaction.open(mmdb, tr_Op)

        # Create the signature
        signum = Signature.populate(tr=tr, subsys=subsys, domain=domain)
        Relvar.insert(mmdb, tr=tr, relvar='Operation_Signature', tuples=[
            Operation_Signature_i(SIGnum=signum, Operation=parsed_op.op, EE=parsed_op.ee, Domain=domain)
        ])

        Relvar.insert(mmdb, tr=tr, relvar='Operation', tuples=[
            Operation_i(Name=parsed_op.op, EE=parsed_op.ee, Domain=domain, Direction=parsed_op.op_type)
        ])
        anum = Activity.populate_operation(tr=tr, action_text=parsed_op.activity,
                                           ee=parsed_op.ee, subsys=subsys, domain=domain,
                                           synchronous=True if parsed_op.flow_out else False)

        if parsed_op.flow_out:
            Relvar.insert(mmdb, tr=tr, relvar='Synchronous_Operation', tuples=[
                Synchronous_Operation_i(Name=parsed_op.op, EE=parsed_op.ee, Domain=domain, Anum=anum)
            ])
        else:
            Relvar.insert(mmdb, tr=tr, relvar='Asynchronous_Operation', tuples=[
                Asynchronous_Operation_i(Name=parsed_op.op, EE=parsed_op.ee, Domain=domain, Anum=anum)
            ])

        Transaction.execute(mmdb, tr)
        logging.info("Transaction closed: Operation")

        # Add parameters
        for p in parsed_op.flows_in:

            # Populate the Parameter's type if it hasn't already been populated
            MMtype.populate_unknown(name=p['type'], domain=domain)
            _logger.info("Transaction open: Populating operation parameter")
            Transaction.open(db=mmdb, name=tr_Parameter)
            input_fid = Flow.populate_data_flow_by_type(mm_type=p['type'], anum=anum,
                                                        domain=domain, label=None, activity_tr=tr_Parameter).fid
            Relvar.insert(db=mmdb, tr=tr_Parameter, relvar='Parameter', tuples=[
                Parameter_i(Name=p['name'], Signature=signum, Domain=domain,
                            Input_flow=input_fid, Activity=anum, Type=p['type'])
            ])
            Transaction.execute(db=mmdb, name=tr_Parameter)
            logging.info("Transaction closed: Parameter")

        # Add output flow
        if parsed_op.flow_out:
            # Populate Synchronous Output and an associated output Data Flow
            Transaction.open(db=mmdb, name=tr_Output)
            of_id = Flow.populate_data_flow_by_type(label=None, mm_type=parsed_op.flow_out,
                                                    anum=anum, domain=domain, activity_tr=tr_Output).fid
            Relvar.insert(db=mmdb, tr=tr_Output, relvar='Synchronous_Output', tuples=[
                Synchronous_Output_i(Anum=anum, Domain=domain,
                                     Output_flow=of_id, Type=parsed_op.flow_out)
            ])
            Transaction.execute(db=mmdb, name=tr_Output)

        return tr_Op



"""
set_action.py – Populate a Set Action instance in PyRAL
"""

import logging
from xuml_populate.config import mmdb
from xuml_populate.populate.actions.table import Table
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.actions.aparse_types import Flow_ap, MaxMult, Content, Activity_ap
from xuml_populate.exceptions.action_exceptions import (ProductForbidsCommonAttributes, UnjoinableHeaders,
                                                        SetOpRequiresSameHeaders)
from xuml_populate.populate.actions.action import Action
from xuml_populate.populate.mm_class import MMclass
from xuml_populate.populate.ns_flow import NonScalarFlow
from xuml_populate.populate.mmclass_nt import Relational_Action_i, Table_Action_i, Set_Action_i
from pyral.relvar import Relvar
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

# Transactions
tr_Set_Action = "Set Action"

class SetAction:
    """
    Create all relations for a ProjectAction
    """
    domain = None
    anum = None
    activity_path = None
    scrall_text = None
    action_id = None
    ns_type = None

    @classmethod
    def populate(cls, a_input: Flow_ap, b_input: Flow_ap, setop: str, activity_data: Activity_ap) -> (str, Flow_ap):
        """
        Populate the Set Action

        :param mmdb: The metamodel database
        :param a_input: The a input Non Scalar Flow
        :param b_input: The b input Non Scalar Flow
        :param setop: The set operation name
        :param activity_data:
        :return: Set action output table flow
        """
        # Save attribute values that we will need when creating the various select subsystem
        # classes
        domain = activity_data.domain
        anum = activity_data.anum

        table_header = None
        max_mult = MaxMult.ONE if a_input.max_mult == b_input.max_mult == MaxMult.ONE else MaxMult.MANY
        match setop:
            case 'JOIN':
                _logger.info("Populating JOIN action")
                # The a/b flows are not joinable if the headers share no common attributes
                if NonScalarFlow.headers_disjoint(a_flow=a_input, b_flow=b_input, domain=domain):
                    raise UnjoinableHeaders
                # There is at least one attribute:type in common, so let's take the union to form the new header
                table_header = NonScalarFlow.header_union(a_flow=a_input, b_flow=b_input, domain=domain)
            case 'UNION' | 'INTERSECT' | 'MINUS':
                _logger.info(f"Populating {setop} action")
                # a/b Types must match (same table or same class)
                if not NonScalarFlow.same_headers(a_input, b_input, domain):
                    raise SetOpRequiresSameHeaders
                # Table header can be taken from a or b since they are the same
                if a_input.content == Content.INSTANCE:
                    table_header = MMclass.header(cname=a_input.tname, domain=domain)
                else:
                    table_header = Table.header(tname=a_input.tname, domain=domain)
            case 'TIMES':
                _logger.info("Populating TIMES action")
                # Verify that there are no attributes in common among the a/b flow
                if not NonScalarFlow.headers_disjoint(a_input, b_input, domain=domain):
                    raise ProductForbidsCommonAttributes
                # Now take the union of the disjoint headers as the output
                table_header = NonScalarFlow.header_union(a_flow=a_input, b_flow=b_input, domain=domain)

        # a/b flow inputs are compatible with the spedified operation
        # Populate the output Table Flow and Table (transaction open/close)
        output_tflow = Flow.populate_relation_flow_by_header(table_header=table_header, anum=anum, domain=domain,
                                                             max_mult=max_mult)

        # Create the action (trannsaction open)
        Transaction.open(db=mmdb, name=tr_Set_Action)
        cls.action_id = Action.populate(tr=tr_Set_Action, anum=anum, domain=domain, action_type="set")
        Relvar.insert(db=mmdb, tr=tr_Set_Action, relvar='Relational_Action', tuples=[
            Relational_Action_i(ID=cls.action_id, Activity=anum, Domain=domain)
        ])
        Relvar.insert(db=mmdb, tr=tr_Set_Action, relvar='Table_Action', tuples=[
            Table_Action_i(ID=cls.action_id, Activity=anum, Domain=domain, Input_a_flow=a_input.fid,
                           Output_flow=output_tflow.fid)
        ])
        Relvar.insert(db=mmdb, tr=tr_Set_Action, relvar='Set_Action', tuples=[
            Set_Action_i(ID=cls.action_id, Operation=setop, Activity=anum, Domain=domain, Input_b_flow=b_input.fid)
        ])
        Transaction.execute(db=mmdb, name=tr_Set_Action)
        return cls.action_id, output_tflow



"""
generalization.py – Process parsed generalization to populate the metamodel db
"""

# System
import logging

# Model Integration
from pyral.relvar import Relvar

# xUML Populate
from xuml_populate.config import mmdb
from xuml_populate.populate.reference import targetid
from xuml_populate.exceptions.mp_exceptions import LessThanTwoSubclassesInGeneralization
from xuml_populate.populate.mmclass_nt import (Generalization_i, Facet_i, Superclass_i, Subclass_i,
                                               Minimal_Partition_i, Reference_i, Generalization_Reference_i,
                                               Formalizing_Class_Role_i, Attribute_Reference_i)

_logger = logging.getLogger(__name__)


class Generalization:
    """
    Populate all relevant Generalization relvars
    """
    def __init__(self, tr: str, domain: str, rnum: str, record):
        """
        Populate a Generalization

        :param tr: The name of the open transaction
        :param domain: The domain name
        :param rnum: The relationship name
        :param record: The relationship parse record
        """
        self.rnum = rnum
        self.domain = domain

        self.subclasses = record['subclasses']
        self.superclass = record['superclass']
        self.genrefs = record['genrefs']
        self.targetid = targetid(ref=self.genrefs[0]['target'], rnum=self.rnum, domain=self.domain)

        # First check for minimal partition
        # TODO We should be using the mmdb to check this constraint, here for now though
        if len(self.subclasses) < 2:
            _logger.error(f"Less than two subclasses in [{self.rnum}].")
            raise LessThanTwoSubclassesInGeneralization(rnum=self.rnum)

        # Populate
        _logger.info(f"Populating Generalization [{self.rnum}]")
        Relvar.insert(db=mmdb, tr=tr, relvar='Generalization', tuples=[
            Generalization_i(Rnum=self.rnum, Domain=domain, Superclass=self.superclass)
        ])
        # Superclass
        Relvar.insert(db=mmdb, tr=tr, relvar='Facet', tuples=[
            Facet_i(Rnum=self.rnum, Domain=domain, Class=self.superclass)
        ])
        Relvar.insert(db=mmdb, tr=tr, relvar='Superclass', tuples=[
            Superclass_i(Rnum=self.rnum, Domain=domain, Class=self.superclass)
        ])
        for subclass in self.subclasses:
            Relvar.insert(db=mmdb, tr=tr, relvar='Facet', tuples=[
                Facet_i(Rnum=self.rnum, Domain=domain, Class=subclass)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Subclass', tuples=[
                Subclass_i(Rnum=self.rnum, Domain=domain, Class=subclass)
            ])
        Relvar.insert(db=mmdb, tr=tr, relvar='Minimal_Partition', tuples=[
            Minimal_Partition_i(Rnum=self.rnum, Domain=domain,
                                A_subclass=self.subclasses[0], B_subclass=self.subclasses[1])
        ])

        # Attribute References
        # If abbreviated, expand <subclass> abbreviation to one explicit reference per subclass
        if len(self.genrefs) == 1 and self.genrefs[0]['source']['class'] == '<subclass>':
            self.genrefs = [{'source': {'class': s, 'attrs': self.genrefs[0]['source']['attrs']},
                            'target': self.genrefs[0]['target'], 'id': self.targetid} for s in self.subclasses]

        for ref in self.genrefs:
            Relvar.insert(db=mmdb, tr=tr, relvar='Reference', tuples=[
                Reference_i(Ref='G',
                            From_class=ref['source']['class'], To_class=ref['target']['class'],
                            Rnum=self.rnum, Domain=domain)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Generalization_Reference', tuples=[
                Generalization_Reference_i(Ref_type='G',
                                           Subclass=ref['source']['class'], Superclass=ref['target']['class'],
                                           Rnum=self.rnum, Domain=domain)
            ])
            Relvar.insert(db=mmdb, tr=tr, relvar='Formalizing_Class_Role', tuples=[
                Formalizing_Class_Role_i(Class=ref['source']['class'], Rnum=self.rnum, Domain=domain)
            ])
            # Create attribute references for each subclass -> superclass reference
            for from_attr, to_attr in zip(ref['source']['attrs'], ref['target']['attrs']):
                Relvar.insert(db=mmdb, tr=tr, relvar='Attribute_Reference', tuples=[
                    Attribute_Reference_i(From_attribute=from_attr, From_class=ref['source']['class'],
                                          To_attribute=to_attr, To_class=ref['target']['class'],
                                          Ref='G',
                                          Domain=domain, To_identifier=self.targetid, Rnum=self.rnum)
                ])



"""
mm_type.py – Pouplate (metamodel) Type instance
"""

import logging
from xuml_populate.config import mmdb
from xuml_populate.populate.mmclass_nt import Type_i, Scalar_i, Table_i, Table_Attribute_i
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction

_logger = logging.getLogger(__name__)

# Transactions
tr_Scalar = "Scalar"
tr_Scalar_Delete = "Scalar Delete"

class MMtype:
    """
    Populate (metamodel) Type instances
    """

    name = None
    domain = None
    mmdb = None
    scalar_types = {}
    class_names = set()

    @classmethod
    def populate_unknown(cls, name: str, domain: str):
        """
        Populate a type that may be Class, Table, or Scalar

        :param name:  The type name
        :param domain: The domain name
        """
        # TODO: For now Table types are not supported
        if name not in cls.class_names:
            cls.populate_scalar(name, domain)

    @classmethod
    def populate_scalar(cls, name: str, domain: str):
        """
        Populate a class type given a class name and domain

        :param name:  Name of a Scalar Type
        :param domain:  Name of its domain
        """
        cls.domain = domain
        cls.name = name

        # Determine if this type has already been defined
        if domain not in cls.scalar_types:
            cls.scalar_types[domain] = set()

        if name in cls.scalar_types[domain]:
            # This type has already been populated
            return

        # Add it to the set of defined scalar types so that we don't populated it more than once
        cls.scalar_types[domain].add(name)

        _logger.info(f"Populating Type for scalar [{cls.name}]")

        Transaction.open(mmdb, tr_Scalar)
        Relvar.insert(mmdb, tr=tr_Scalar, relvar='Type', tuples=[
            Type_i(Name=cls.name, Domain=cls.domain)
        ])
        Relvar.insert(mmdb, tr=tr_Scalar, relvar='Scalar', tuples=[
            Scalar_i(Name=cls.name, Domain=cls.domain)
        ])
        Transaction.execute(mmdb, tr_Scalar)


    @classmethod
    def populate_class(cls, tr: str, cname: str, domain: str):
        """
        Populate a class type given a class name and domain

        :param mmdb: The metamodel db name
        :param tr:  The name of the open db transaction
        :param cname:  Name of some class
        :param domain:  Name of its domain
        """
        cls.domain = domain
        cls.name = cname

        cls.class_names.add(cname)

        _logger.info(f"Populating Type for class [{cls.name}]")
        Relvar.insert(mmdb, tr=tr, relvar='Type', tuples=[
            Type_i(Name=cls.name, Domain=cls.domain)
        ])

    @classmethod
    def depopulate_scalar(cls, name: str, domain: str):
        """
        Remove the specified Scalar from the database.

        The only use case for this currently is the removal of the dummy UNRESOLVED Scalar

        :param mmdb: The metamodel db name
        :param name:  The Scalar name
        :param domain:  The domain name
        :return:
        """
        # Verify that the scalar exists
        R = f"Name:<{name}>, Domain:<{domain}>"
        result = Relation.restrict(mmdb, restriction=R, relation="Type").body
        if not result:
            # TODO: This is happening for some reason, but doesn't seem to be an error
            _logger.error("Scalar dummy UNRESOLVED not found during depopulate")
        # Depopulate scalar
        Transaction.open(mmdb, tr_Scalar_Delete)
        Relvar.deleteone(mmdb, tr=tr_Scalar_Delete, relvar_name='Type', tid={'Name': name, 'Domain': domain})
        Relvar.deleteone(mmdb, tr=tr_Scalar_Delete, relvar_name='Scalar', tid={'Name': name, 'Domain': domain})
        Transaction.execute(mmdb, tr_Scalar_Delete)



"""
domain.py – Process parsed domain to populate the metamodel db
"""

# System
import logging
from typing import Dict

# Model Integration
from pyral.transaction import Transaction
from pyral.relvar import Relvar
from pyral.relvar import Relation  # For debugging

# xUML Populate
from xuml_populate.config import mmdb
from xuml_populate.populate.attribute import Attribute
from xuml_populate.populate.mm_class import MMclass
from xuml_populate.populate.method import Method
from xuml_populate.populate.relationship import Relationship
from xuml_populate.populate.ee import EE
from xuml_populate.populate.lineage import Lineage
from xuml_populate.populate.subsystem import Subsystem
from xuml_populate.populate.state_model import StateModel
from xuml_populate.populate.activity import Activity
from xuml_populate.populate.mmclass_nt import Domain_i, Modeled_Domain_i, Domain_Partition_i, Subsystem_i


_logger = logging.getLogger(__name__)

# Transactions
tr_Modeled_Domain = "Modeled Domain"

class Domain:
    """
    Populate all relevant Domain relvars
    """
    def __init__(self, domain: str, content: Dict, parse_actions: bool, verbose: bool):
        """
        Insert all user model elements in this Domain into the corresponding Metamodel classes.

        :param domain:  The name of the domain extracted from the content
        :param content:  The parsed content of the domain
        """
        _logger.info(f"Populating modeled domain [{domain}]")

        self.subsystem_counter = {}
        self.types = None
        self.parse_actions = parse_actions

        _logger.info(f"Transaction open: domain and subsystems [{domain}]")
        Transaction.open(db=mmdb, name=tr_Modeled_Domain)

        Relvar.insert(db=mmdb, tr=tr_Modeled_Domain, relvar='Domain', tuples=[
            Domain_i(Name=domain, Alias=content['alias']),
        ])
        # TODO: For now assume this is always a modeled domain, but need a way to specify a realized domain
        Relvar.insert(db=mmdb, tr=tr_Modeled_Domain, relvar='Modeled_Domain', tuples=[
            Modeled_Domain_i(Name=domain),
            ])
        for subsys_parse in content['subsystems'].values():
            subsys = subsys_parse['class_model'].subsystem
            Relvar.insert(db=mmdb, tr=tr_Modeled_Domain, relvar='Subsystem', tuples=[
                Subsystem_i(Name=subsys['name'], First_element_number=subsys['range'][0],
                            Domain=domain, Alias=subsys['alias']),
            ])
            Relvar.insert(db=mmdb, tr=tr_Modeled_Domain, relvar='Domain_Partition', tuples=[
                Domain_Partition_i(Number=subsys['range'][0], Domain=domain)
            ])
        Transaction.execute(db=mmdb, name=tr_Modeled_Domain)
        _logger.info(f"Transaction closed: domain and subsystems [{domain}]")

        # Process all subsystem elements
        for subsys_parse in content['subsystems'].values():
            subsys = Subsystem(subsys_parse=subsys_parse['class_model'].subsystem)
            _logger.info("Populating classes")

            # Insert classes
            for c in subsys_parse['class_model'].classes:
                MMclass.populate(domain=domain, subsystem=subsys, record=c)
            _logger.info("Populating relationships")
            for r in subsys_parse['class_model'].rels:
                Relationship.populate(domain=domain, subsystem=subsys, record=r)

            # Insert external entities and operations
            _logger.info("Populating operations")
            for ee_name, op_parse in subsys_parse['external'].items():
                EE.populate(ee_name=ee_name, subsys=subsys.name,
                            domain=domain, op_parse=op_parse)

            # Insert methods
            _logger.info("Populating methods")
            for m_parse in subsys_parse['methods'].values():
                # All classes must be populated first, so that parameter types in signatures can be resolved
                # as class or non-class types
                Method(domain=domain, subsys=subsys.name, m_parse=m_parse, parse_actions=parse_actions)

            # Insert state machines
            _logger.info("Populating state models")
            for sm in subsys_parse['state_models'].values():
                StateModel(subsys=subsys.name, sm=sm, parse_actions=self.parse_actions)

        Attribute.ResolveAttrTypes(domain=domain)
        _logger.info("Populating lineage")
        #
        # Reprinting these for lineage debugging purposes
        Lineage.Derive(domain=domain)
        #
        #
        # Populate actions for all Activities
        Activity.process_execution_units()  # TODO: For now, only methods
        # if self.parse_actions:
        #     Activity.process_execution_units()

        # Print out the populated metamodel
        if verbose:
            Relvar.printall(mmdb)
        #



"""
state_model.py – Process parsed lifecycle to p
"""

# System
import logging

# Model Integration
from xsm_parser.state_model_parser import StateModel_a
from pyral.relvar import Relvar
from pyral.relation import Relation
from pyral.transaction import Transaction


# xUML Populate
from xuml_populate.config import mmdb
from xuml_populate.exceptions.mp_exceptions import MismatchedStateSignature, BadStateModelName
from xuml_populate.populate.flow import Flow
from xuml_populate.populate.signature import Signature
from xuml_populate.populate.activity import Activity
from xuml_populate.populate.mm_type import MMtype
from xuml_populate.populate.mmclass_nt import (State_Model_i, Lifecycle_i, Non_Deletion_State_i, State_i, Real_State_i,
                                               Deletion_State_i, Initial_Pseudo_State_i, State_Signature_i,
                                               Initial_Transition_i, Event_Response_i, Transition_i, Non_Transition_i,
                                               Event_Specification_i, Monomorphic_Event_Specification_i,
                                               Monomorphic_Event_i, Effective_Event_i, Event_i, Parameter_i, Assigner_i,
                                               Multiple_Assigner_i, Single_Assigner_i)

ISP = 'Initial_Pseudo_State'  # Name of initial pseudo state

_logger = logging.getLogger(__name__)
tr_SM = 'State Model'


class StateModel:
    """
    Create a State Model relation
    """

    def __init__(self, subsys: str, sm: StateModel_a, parse_actions: bool):
        """Constructor"""

        self.cname = sm.lifecycle
        self.rnum = sm.assigner_rnum
        # Ensure that only cname or rnum is set
        if not self.cname and not self.rnum:
            msg = f"No state model rnum or class name specified"
            _logger.exception(msg)
            raise BadStateModelName(msg)
        if self.cname and self.rnum:
            msg = f"State model must be either lifecycle (cname) or assigner (rnum), not both"
            _logger.exception(msg)
            raise BadStateModelName(msg)
        self.pclass = sm.assigner_pclass  # Establishes a multiple assigner with a partitioning class
        self.sm_name = self.cname if self.cname else self.rnum
        self.signums = {}  # Remember signature of each inserted state for processing transitions
        self.signatures = {}
        self.parse_actions = parse_actions

        # Populate
        # It is easiest to create all events and states at once before checking constraints
        Transaction.open(db=mmdb, name=tr_SM)
        Relvar.insert(db=mmdb, tr=tr_SM, relvar='State_Model', tuples=[
            State_Model_i(Name=self.sm_name, Domain=sm.domain)
        ])
        if self.cname:  # Lifecycle state model
            _logger.info(f"Populating Lifecycle [{self.cname}]")
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Lifecycle', tuples=[
                Lifecycle_i(Class=self.cname, Domain=sm.domain)
            ])
        else:  # Assigner state model
            _logger.info(f"Populating Assigner [{self.rnum}]")
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Assigner', tuples=[
                Assigner_i(Rnum=self.rnum, Domain=sm.domain)
            ])
            if self.pclass:
                # Multiple assigner with a partitioning class
                Relvar.insert(db=mmdb, tr=tr_SM, relvar='Multiple_Assigner', tuples=[
                    Multiple_Assigner_i(Rnum=self.rnum, Partitioning_class=self.pclass, Domain=sm.domain)
                ])
            else:
                # Single assigner
                Relvar.insert(db=mmdb, tr=tr_SM, relvar='Single_Assigner', tuples=[
                    Single_Assigner_i(Rnum=self.rnum, Domain=sm.domain)
                ])

        # Populate the states
        for s in sm.states:
            # Create Real State and all associated model elements
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='State', tuples=[
                State_i(Name=s.state.name, State_model=self.sm_name, Domain=sm.domain)
            ])
            # Create its Activity
            anum = Activity.populate_state(tr=tr_SM,
                                           state=s.state.name, subsys=subsys, actions=s.activity,
                                           state_model=self.sm_name, domain=sm.domain,
                                           parse_actions=self.parse_actions
                                           )
            # Signature
            sig_params = frozenset(s.state.signature)
            if sig_params not in self.signatures:
                # Add new signature if it doesn't exist
                # First create signature superclass instance in Activity subsystem
                signum = Signature.populate(tr=tr_SM, subsys=subsys, domain=sm.domain)
                self.signatures[sig_params] = signum  # Save the SIGnum as a value, keyed to the frozen params
                Relvar.insert(db=mmdb, tr=tr_SM, relvar='State_Signature', tuples=[
                    State_Signature_i(SIGnum=signum, State_model=self.sm_name, Domain=sm.domain)
                ])
                # Now we need to create Data Flows and Parameters
                for p in s.state.signature:
                    # Create a Data flow
                    # Populate the Parameter's type if it hasn't already been populated
                    MMtype.populate_unknown(name=p.type, domain=sm.domain)
                    input_fid = Flow.populate_data_flow_by_type(mm_type=p.type, anum=anum,
                                                                domain=sm.domain, label=None,
                                                                activity_tr=tr_SM).fid
                    Relvar.insert(db=mmdb, tr=tr_SM, relvar='Parameter', tuples=[
                        Parameter_i(Name=p.name, Signature=signum, Domain=sm.domain,
                                    Input_flow=input_fid, Activity=anum, Type=p.type)
                    ])
            else:
                # Otherwise, just get the id of the matching signature
                signum = self.signatures[sig_params]
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Real_State', tuples=[
                Real_State_i(Name=s.state.name, State_model=self.sm_name, Domain=sm.domain, Signature=signum,
                             Activity=anum)
            ])
            # We need to look up the sid when matching events on incoming transitions
            self.signums[s.state.name] = signum
            if self.rnum or self.cname and not s.state.deletion:
                # Assigner cannot have a Deletion State
                Relvar.insert(db=mmdb, tr=tr_SM, relvar='Non_Deletion_State', tuples=[
                    Non_Deletion_State_i(Name=s.state.name, State_model=self.sm_name, Domain=sm.domain)
                ])
            else:
                Relvar.insert(db=mmdb, tr=tr_SM, relvar='Deletion_State', tuples=[
                    Deletion_State_i(Name=s.state.name, Class=self.cname, Domain=sm.domain)
                ])

        # Populate the events
        # TODO: Handle polymorphic events
        for ev_name in sm.events:
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Event', tuples=[
                Event_i(Name=ev_name, State_model=self.sm_name, Domain=sm.domain)
            ])
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Effective_Event', tuples=[
                Effective_Event_i(Name=ev_name, State_model=self.sm_name, Domain=sm.domain)
            ])
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Monomorphic_Event', tuples=[
                Monomorphic_Event_i(Name=ev_name, State_model=self.sm_name, Domain=sm.domain)
            ])
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Monomorphic_Event_Specification', tuples=[
                Monomorphic_Event_Specification_i(Name=ev_name, State_model=self.sm_name, Domain=sm.domain)
            ])
            # Cannot create Event Specification until we process transitions to determine signature for at least
            # one target state

        # Populate the transitions
        inserted_especs = {}
        if sm.initial_transitions:
            # Create the intial pseudo state
            Relvar.insert(db=mmdb, tr=tr_SM, relvar=ISP, tuples=[
                Initial_Pseudo_State_i(Name=ISP, Class=self.sm_name, Domain=sm.domain)
            ])
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='State', tuples=[
                State_i(Name=ISP, State_model=self.sm_name, Domain=sm.domain)
            ])
        for t in sm.initial_transitions:
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Initial_Transition', tuples=[
                Initial_Transition_i(From_state=ISP, Class=self.sm_name, Domain=sm.domain, Event=t.event)
            ])
            signum = self.signums[t.to_state]
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Event_Specification', tuples=[
                Event_Specification_i(Name=t.event, State_model=self.sm_name, Domain=sm.domain,
                                      State_signature=signum)
            ])
            inserted_especs[t.event] = signum
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Event_Response', tuples=[
                Event_Response_i(State=ISP, Event=t.event, State_model=self.sm_name, Domain=sm.domain)
            ])
            Relvar.insert(db=mmdb, tr=tr_SM, relvar='Transition', tuples=[
                Transition_i(From_state=ISP, Event=t.event, State_model=self.sm_name, Domain=sm.domain,
                             To_state=t.to_state)
            ])
            for e in sm.events:
                if e != t.event:
                    Relvar.insert(db=mmdb, tr=tr_SM, relvar='Event_Response', tuples=[
                        Event_Response_i(State=ISP, Event=e, State_model=self.sm_name, Domain=sm.domain)
                    ])
                    Relvar.insert(db=mmdb, tr=tr_SM, relvar='Non_Transition', tuples=[
                        Non_Transition_i(State=ISP, Event=e, State_model=self.sm_name, Domain=sm.domain,
                                         Behavior='CH',
                                         Reason="Transition is only legal response from initial pseudo state")
                    ])

        for s in sm.states:
            if s.state.deletion:
                break  # There are no transitions out of a deletion state
            for t in s.transitions:
                if t.to_state:
                    # Insert or check event spec signature
                    if t.event not in inserted_especs:
                        # The event spec will assume the signature of the first target state encountered
                        signum = self.signums[t.to_state]
                        Relvar.insert(db=mmdb, tr=tr_SM, relvar='Event_Specification', tuples=[
                            Event_Specification_i(Name=t.event, State_model=self.sm_name, Domain=sm.domain,
                                                  State_signature=signum)
                        ])
                        inserted_especs[t.event] = signum  # Remember for matching in the else clause below
                    else:
                        # We need to verify that the to_state's signature matches that of the event spec
                        state_sig = self.signums[t.to_state]
                        espec_sig = inserted_especs[t.event]
                        if state_sig != espec_sig:
                            _logger.error(
                                f"Mismatched espec sig: <{t.event}:{espec_sig}> state sig: [{t.to_state}:{state_sig}]")
                            raise MismatchedStateSignature(event=t.event, state=t.to_state)
                    # Create transition event response
                    Relvar.insert(db=mmdb, tr=tr_SM, relvar='Event_Response', tuples=[
                        Event_Response_i(State=s.state.name, Event=t.event, State_model=self.sm_name, Domain=sm.domain)
                    ])
                    Relvar.insert(db=mmdb, tr=tr_SM, relvar='Transition', tuples=[
                        Transition_i(From_state=s.state.name, Event=t.event, State_model=self.sm_name, Domain=sm.domain,
                                     To_state=t.to_state)
                    ])
                else:  # Create Non Transition ignore response
                    Relvar.insert(db=mmdb, tr=tr_SM, relvar='Event_Response', tuples=[
                        Event_Response_i(State=s.state.name, Event=t.event, State_model=self.sm_name, Domain=sm.domain)
                    ])
                    Relvar.insert(db=mmdb, tr=tr_SM, relvar='Non_Transition', tuples=[
                        Non_Transition_i(State=s.state.name, Event=t.event, State_model=self.sm_name, Domain=sm.domain,
                                         Behavior='IGN', Reason="<none_specified>")
                    ])

        # All event specs have been created, now fill in the can't happens
        for s in sm.states:
            tr_ign_events = set(t.event for t in s.transitions)
            for e in sm.events:
                if e not in tr_ign_events:
                    Relvar.insert(db=mmdb, tr=tr_SM, relvar='Event_Response', tuples=[
                        Event_Response_i(State=s.state.name, Event=e, State_model=self.sm_name, Domain=sm.domain)
                    ])
                    ch_reason = "<none_specified>" if not s.state.deletion else "Event cannot happen in deletion state"
                    Relvar.insert(db=mmdb, tr=tr_SM, relvar='Non_Transition', tuples=[
                        Non_Transition_i(State=s.state.name, Event=e, State_model=self.sm_name, Domain=sm.domain,
                                         Behavior='CH', Reason=ch_reason)
                    ])

        Transaction.execute(db=mmdb, name=tr_SM)



"""
anum.py – Populate an Activity
"""

# System
import logging
from typing import NamedTuple

# Model Integration
from pyral.relvar import Relvar
from pyral.relation import Relation
from scrall.parse.parser import ScrallParser

# xUML Populate
from xuml_populate.config import mmdb
from xuml_populate.populate.method_activity import MethodActivity
from xuml_populate.populate.element import Element
from xuml_populate.populate.actions.aparse_types import Activity_ap
from xuml_populate.populate.mmclass_nt import (Activity_i, Asynchronous_Activity_i, State_Activity_i,
                                               Synchronous_Activity_i)

_logger = logging.getLogger(__name__)

# Temporarily silence the noisy scrall visitor logger
null_handler = logging.NullHandler()
# print(logging.root.manager.loggerDict.keys())
slogger = logging.getLogger("scrall.parse.visitor")
slogger.handlers.clear()
slogger.addHandler(null_handler)
slogger.propagate = False


class Activity:
    """
    Create an Activity relation
    """
    # These dictionaries are for debugging purposes, delete once we get action semantics populated
    sm = {}
    methods = {}
    operations = {}
    domain = None

    @classmethod
    def valid_param(cls, pname: str, activity: Activity_ap):
        # TODO: Verify that the parameter is in the signature of the specified activity with exception if not
        pass

    @classmethod
    def populate_method(cls, tr: str, action_text: str, cname: str, method: str,
                        subsys: str, domain: str, parse_actions: bool) -> str:
        """
        Populate Synchronous Activity for Method

        :param tr: The name of the open transaction
        :param cname: The class name
        :param method: The method name
        :param action_text: Unparsed scrall text
        :param subsys: The subsystem name
        :param domain: The domain name
        :param parse_actions:
        :return: The Activity number (Anum)
        """
        cls.domain = domain
        Anum = cls.populate(tr=tr, action_text=action_text, subsys=subsys, domain=domain, synchronous=True)
        if cname not in cls.methods:
            cls.methods[cname] = {
                method: {'anum': Anum, 'domain': domain, 'text': action_text, 'parse': None}}
        else:
            cls.methods[cname][method] = {'anum': Anum, 'domain': domain, 'text': action_text, 'parse': None}
        # Parse the scrall and save for later population
        cls.methods[cname][method]['parse'] = ScrallParser.parse_text(scrall_text=action_text, debug=False)
        pass
        # if parse_actions:
        #     cls.methods[cname][method]['parse'] = ScrallParser.parse_text(scrall_text=action_text, debug=False)
        # else:
        #     cls.methods[cname][method]['parse'] = None
        return Anum

    @classmethod
    def populate_operation(cls, tr: str, action_text: str, ee: str, subsys: str, domain: str,
                           synchronous: bool) -> str:
        """
        Populate Operation Activity

        :param tr:
        :param action_text:
        :param ee:
        :param subsys:
        :param domain:
        :param synchronous:
        :return:
        """
        Anum = cls.populate(tr=tr, action_text=action_text, subsys=subsys, domain=domain, synchronous=synchronous)
        cls.operations[ee] = ScrallParser.parse_text(scrall_text=action_text, debug=False)
        return Anum

    @classmethod
    def populate(cls, tr: str, action_text: str, subsys: str, domain: str,
                 synchronous: bool) -> str:
        """
        Populate an Activity

        :param tr: The name of the open transaction
        :param action_text: Unparsed scrall text
        :param subsys: The subsystem name
        :param domain: The domain name
        :param synchronous: True if Activity is synchronous
        :return: The Activity number (Anum)
        """
        Anum = Element.populate_unlabeled_subsys_element(tr=tr, prefix='A', subsystem=subsys, domain=domain)
        Relvar.insert(db=mmdb, tr=tr, relvar='Activity', tuples=[
            Activity_i(Anum=Anum, Domain=domain)
        ])
        if synchronous:
            Relvar.insert(db=mmdb, tr=tr, relvar='Synchronous_Activity', tuples=[
                Synchronous_Activity_i(Anum=Anum, Domain=domain)
            ])
        else:
            Relvar.insert(db=mmdb, tr=tr, relvar='Asynchronous_Activity', tuples=[
                Asynchronous_Activity_i(Anum=Anum, Domain=domain)
            ])
        return Anum

    @classmethod
    def populate_state(cls, tr: str, state: str, state_model: str, actions: str,
                       subsys: str, domain: str, parse_actions: bool) -> str:
        """
        :param tr:  Name of the transaction
        :param state: State name
        :param state_model:  State model name
        :param actions:
        :param subsys:
        :param domain:
        :param parse_actions:
        :return: Anum
        """

        # Parse scrall in this state and add it to temporary sm dictionary
        action_text = ''.join(actions) + '\n'
        if state_model not in cls.sm:
            cls.sm[state_model] = {}
        if parse_actions:
            parsed_activity = ScrallParser.parse_text(scrall_text=action_text, debug=False)
        else:
            parsed_activity = None
        cls.sm[state_model][state] = parsed_activity  # To subsys_parse parsed actions for debugging
        # cls.populate_activity(text=action_text, pa=parsed_activity)

        # Create the Susbystem Element and obtain a unique Anum
        Anum = cls.populate(tr=tr, action_text=action_text, subsys=subsys, domain=domain, synchronous=False)
        Relvar.insert(db=mmdb, tr=tr, relvar='State_Activity', tuples=[
            State_Activity_i(Anum=Anum, State=state, State_model=state_model, Domain=domain)
        ])
        return Anum


    @classmethod
    def process_execution_units(cls):
        """
        Process each Scrall Execution Unit for all Activities (Method, State, and Synchronous Operation)
        """
        # Populate each (Method) Activity
        for class_name, method_data in cls.methods.items():
            for method_name, activity_data in method_data.items():

                ma = MethodActivity(name=method_name, class_name=class_name, method_data=method_data,
                                    activity_data=activity_data, domain=cls.domain)
        pass

    # TODO: Populate all state activities
    # TODO: Populate all operation activities
